import { Observer } from '../runtime';
export declare class ComputedTracker {
    static isTracking: boolean;
    static trackedObservers: Set<Observer>;
    /**
     * Helper Class for automatic tracking used Observers (dependencies) in a compute function.
     *
     * @internal
     */
    constructor();
    /**
     * Activates the Computed Tracker to globally track used Observers.
     *
     * @internal
     */
    static track(): void;
    /**
     * Tracks the specified Observer and caches it
     * if the Computed Tracker is actively tracking.
     *
     * @internal
     * @param observer - Observer
     */
    static tracked(observer: Observer): void;
    /**
     * Returns the latest tracked Observers
     * and stops the Computed Tracker from tracking any more Observers.
     *
     * @internal
     */
    static getTrackedObservers(): Array<Observer>;
}
