import { defineConfig } from '@agile-ts/utils';
import { logCodeManager } from './logCodeManager.js';
import { globalBind } from './utils.js';
import { Integrations } from './integrations/integrations.js';
import { Runtime } from './runtime/runtime.js';
import { SubController } from './runtime/subscription/sub.controller.js';

const _Agile = class {
  constructor(config = {}) {
    config = defineConfig(config, {
      localStorage: false,
      waitForMount: true,
      bindGlobal: false,
      autoIntegrate: true,
      bucket: true
    });
    this.config = {
      waitForMount: config.waitForMount,
      bucket: config.bucket
    };
    this.key = config.key;
    this.integrations = new Integrations(this, {
      autoIntegrate: config.autoIntegrate
    });
    this.runtime = new Runtime(this);
    this.subController = new SubController(this);
    logCodeManager.log("10:00:00", {}, this);
    if (config.bindGlobal) {
      if (!globalBind(_Agile.globalKey, this)) {
        logCodeManager.log("10:02:00");
      }
    }
  }
  integrate(integration) {
    this.integrations.integrate(integration);
    return this;
  }
  hasIntegration() {
    return this.integrations.hasIntegration();
  }
};
let Agile = _Agile;
Agile.globalKey = "__agile__";

export { Agile };
