import { defineConfig, copy, generateId, equal } from '@agile-ts/utils';
import { logCodeManager } from '../../logCodeManager.js';
import { Observer } from '../../runtime/observer.js';
import { RuntimeJob } from '../../runtime/runtime.job.js';

class GroupObserver extends Observer {
  constructor(group, config = {}) {
    super(group.agileInstance(), defineConfig(config, {
      value: copy(group._output)
    }));
    this.group = () => group;
    this.nextGroupValue = copy(group._output);
  }
  ingest(config = {}) {
    this.ingestOutput(this.group().nextGroupOutput, config);
  }
  ingestOutput(newGroupOutput, config = {}) {
    const group = this.group();
    config = defineConfig(config, {
      perform: true,
      force: false,
      key: logCodeManager.allowLogging ? `${this.key != null ? this.key + "_" : ""}${generateId()}_output` : void 0
    });
    if (group.isPlaceholder) {
      config.force = true;
    }
    this.nextGroupValue = copy(newGroupOutput);
    if (equal(group._output, this.nextGroupValue) && !config.force)
      return;
    const job = new RuntimeJob(this, config);
    this.agileInstance().runtime.ingest(job, {
      perform: config.perform
    });
  }
  perform(job) {
    const observer = job.observer;
    const group = observer.group();
    group._output = copy(observer.nextGroupValue);
    group.nextGroupOutput = copy(observer.nextGroupValue);
    observer.previousValue = Object.freeze(copy(observer.value));
    observer.value = copy(group._output);
  }
}

export { GroupObserver };
