import { defineConfig } from '@agile-ts/utils';
import { CollectionPersistent } from './collection.persistent.js';
import { EnhancedState } from '../state/state.enhanced.js';

const _Item = class extends EnhancedState {
  constructor(collection, data, config = {}) {
    super(collection.agileInstance(), data, {
      isPlaceholder: config.isPlaceholder,
      key: data[collection.config.primaryKey]
    });
    this.selectedBy = new Set();
    this.collection = () => collection;
    if (this._key != null) {
      this.addRebuildGroupThatIncludeItemKeySideEffect(this._key);
    }
  }
  setKey(value, config = {}) {
    super.setKey(value);
    config = defineConfig(config, {
      sideEffects: {
        enabled: true,
        exclude: []
      },
      background: false,
      force: false,
      storage: true,
      overwrite: false
    });
    if (value == null)
      return this;
    this.removeSideEffect(_Item.updateGroupSideEffectKey);
    this.addRebuildGroupThatIncludeItemKeySideEffect(value);
    this.patch({ [this.collection().config.primaryKey]: value }, config);
    return this;
  }
  persist(config = {}) {
    config = defineConfig(config, {
      key: this._key,
      followCollectionPersistKeyPattern: true
    });
    if (config.followCollectionPersistKeyPattern) {
      config.key = CollectionPersistent.getItemStorageKey(config.key || this._key, this.collection()._key);
    }
    super.persist(config);
    return this;
  }
  addRebuildGroupThatIncludeItemKeySideEffect(itemKey) {
    this.addSideEffect(_Item.updateGroupSideEffectKey, (instance, config) => {
      instance.collection().rebuildGroupsThatIncludeItemKey(itemKey, config);
    }, { weight: 100 });
  }
};
let Item = _Item;
Item.updateGroupSideEffectKey = "rebuildGroup";

export { Item };
