import { defineConfig } from '@agile-ts/utils';
import { logCodeManager } from '../logCodeManager.js';
import { Integration } from './integration.js';

typeof require !== "undefined" ? require : (x) => {
  throw new Error('Dynamic require of "' + x + '" is not supported');
};
var __async = (__this, __arguments, generator) => {
  return new Promise((resolve, reject) => {
    var fulfilled = (value) => {
      try {
        step(generator.next(value));
      } catch (e) {
        reject(e);
      }
    };
    var rejected = (value) => {
      try {
        step(generator.throw(value));
      } catch (e) {
        reject(e);
      }
    };
    var step = (x) => x.done ? resolve(x.value) : Promise.resolve(x.value).then(fulfilled, rejected);
    step((generator = generator.apply(__this, __arguments)).next());
  });
};
const onRegisterInitialIntegrationCallbacks = [];
const _Integrations = class {
  constructor(agileInstance, config = {}) {
    this.integrations = new Set();
    config = defineConfig(config, {
      autoIntegrate: true
    });
    this.agileInstance = () => agileInstance;
    if (config.autoIntegrate) {
      _Integrations.onRegisterInitialIntegration((integration) => {
        this.integrate(integration);
      });
    }
  }
  static addInitialIntegration(integration) {
    if (integration instanceof Integration) {
      onRegisterInitialIntegrationCallbacks.forEach((callback) => callback(integration));
      _Integrations.initialIntegrations.push(integration);
    }
  }
  static onRegisterInitialIntegration(callback) {
    onRegisterInitialIntegrationCallbacks.push(callback);
    _Integrations.initialIntegrations.forEach((integration) => {
      callback(integration);
    });
  }
  integrate(integration) {
    return __async(this, null, function* () {
      if (integration.key == null) {
        if (process.env.NODE_ENV !== "production") {
          logCodeManager.log("18:03:00", { replacers: [integration.key, this.agileInstance().key] }, integration);
        }
        return false;
      }
      if (integration.methods.bind)
        integration.ready = yield integration.methods.bind(this.agileInstance());
      else
        integration.ready = true;
      this.integrations.add(integration);
      integration.integrated = true;
      if (process.env.NODE_ENV !== "production") {
        logCodeManager.log("18:00:00", { replacers: [integration.key, this.agileInstance().key] }, integration);
      }
      return true;
    });
  }
  update(componentInstance, updatedData) {
    this.integrations.forEach((integration) => {
      if (!integration.ready) {
        logCodeManager.log("18:02:00", { replacers: [integration.key] });
        return;
      }
      if (integration.methods.updateMethod)
        integration.methods.updateMethod(componentInstance, updatedData);
    });
  }
  hasIntegration() {
    return this.integrations.size > 0;
  }
};
let Integrations = _Integrations;
Integrations.initialIntegrations = [];

export { Integrations };
