import { defineConfig, generateId } from '@agile-ts/utils';
import { logCodeManager } from '../logCodeManager.js';
import { RuntimeJob } from './runtime.job.js';

class Observer {
  constructor(agileInstance, config = {}) {
    this.dependents = new Set();
    this.subscribedTo = new Set();
    var _a, _b;
    config = defineConfig(config, {
      dependents: [],
      subs: [],
      value: null
    });
    this.agileInstance = () => agileInstance;
    this.key = config.key;
    this.value = config.value;
    this.previousValue = config.value;
    (_a = config.dependents) == null ? void 0 : _a.forEach((observer) => this.addDependent(observer));
    (_b = config.subs) == null ? void 0 : _b.forEach((subscriptionContainer) => subscriptionContainer.addSubscription(this));
  }
  ingest(config = {}) {
    config = defineConfig(config, {
      perform: true,
      key: logCodeManager.allowLogging ? `${this.key != null ? this.key + "_" : ""}${generateId()}` : void 0
    });
    const job = new RuntimeJob(this, config);
    this.agileInstance().runtime.ingest(job, {
      perform: config.perform
    });
  }
  perform(job) {
    if (process.env.NODE_ENV !== "production") {
      logCodeManager.log("17:03:00");
    }
  }
  addDependent(observer) {
    if (!this.dependents.has(observer))
      this.dependents.add(observer);
  }
  removeDependent(observer) {
    if (this.dependents.has(observer))
      this.dependents.delete(observer);
  }
}

export { Observer };
