import { defineConfig, generateId, isValidObject } from '@agile-ts/utils';

class SubscriptionContainer {
  constructor(subs, config = {}) {
    this.ready = false;
    this.updatedSubscribers = new Set();
    this.isObjectBased = false;
    var _a, _b, _c, _d;
    config = defineConfig(config, {
      key: generateId()
    });
    this.subscribers = new Set();
    this.key = config.key;
    this.componentId = config == null ? void 0 : config.componentId;
    this.subscriberKeysWeakMap = new WeakMap();
    this.selectorsWeakMap = new WeakMap();
    this.isObjectBased = !Array.isArray(subs);
    for (const key in subs) {
      this.addSubscription(subs[key], {
        proxyPaths: (_b = (_a = config.proxyWeakMap) == null ? void 0 : _a.get(subs[key])) == null ? void 0 : _b.paths,
        selectorMethods: (_d = (_c = config.selectorWeakMap) == null ? void 0 : _c.get(subs[key])) == null ? void 0 : _d.methods,
        key: this.isObjectBased ? key : void 0
      });
    }
  }
  addSubscription(sub, config = {}) {
    var _a, _b, _c, _d;
    const toAddSelectorMethods = (_a = config.selectorMethods) != null ? _a : [];
    const proxyPaths = (_b = config.proxyPaths) != null ? _b : [];
    for (const path of proxyPaths) {
      toAddSelectorMethods.push((value) => {
        let _value = value;
        for (const branch of path) {
          if (!isValidObject(_value, true))
            break;
          _value = _value[branch];
        }
        return _value;
      });
    }
    const existingSelectorMethods = (_d = (_c = this.selectorsWeakMap.get(sub)) == null ? void 0 : _c.methods) != null ? _d : [];
    const newSelectorMethods = existingSelectorMethods.concat(toAddSelectorMethods);
    if (newSelectorMethods.length > 0)
      this.selectorsWeakMap.set(sub, { methods: newSelectorMethods });
    if (config.key != null)
      this.subscriberKeysWeakMap.set(sub, config.key);
    this.subscribers.add(sub);
    sub.subscribedTo.add(this);
  }
  removeSubscription(sub) {
    if (this.subscribers.has(sub)) {
      this.selectorsWeakMap.delete(sub);
      this.subscriberKeysWeakMap.delete(sub);
      this.subscribers.delete(sub);
      sub.subscribedTo.delete(this);
    }
  }
}

export { SubscriptionContainer };
