import { defineConfig, isAsyncFunction, isFunction, isJsonString } from '@agile-ts/utils';
import { logCodeManager } from '../logCodeManager.js';

class Storage {
  constructor(config) {
    this.ready = false;
    config = defineConfig(config, {
      prefix: "agile",
      async: false
    });
    this.key = config.key;
    this.methods = config.methods;
    this.config = {
      prefix: config.prefix,
      async: config.async
    };
    this.ready = this.validate();
    if (!this.ready)
      return;
    if (isAsyncFunction(this.methods.get) || isAsyncFunction(this.methods.set) || isAsyncFunction(this.methods.remove))
      this.config.async = true;
  }
  validate() {
    var _a, _b, _c;
    if (!isFunction((_a = this.methods) == null ? void 0 : _a.get)) {
      logCodeManager.log("13:03:00", { replacers: ["get"] });
      return false;
    }
    if (!isFunction((_b = this.methods) == null ? void 0 : _b.set)) {
      logCodeManager.log("13:03:00", { replacers: ["set"] });
      return false;
    }
    if (!isFunction((_c = this.methods) == null ? void 0 : _c.remove)) {
      logCodeManager.log("13:03:00", { replacers: ["remove"] });
      return false;
    }
    return true;
  }
  normalGet(key) {
    if (!this.ready || !this.methods.get)
      return void 0;
    if (isAsyncFunction(this.methods.get))
      logCodeManager.log("13:02:00");
    const res = this.methods.get(this.getStorageKey(key));
    const _res = isJsonString(res) ? JSON.parse(res) : res;
    logCodeManager.log("13:01:00", { tags: ["storage"], replacers: [this.key, this.getStorageKey(key)] }, _res);
    return _res;
  }
  get(key) {
    if (!this.ready || !this.methods.get)
      return Promise.resolve(void 0);
    if (!isAsyncFunction(this.methods.get))
      return Promise.resolve(this.normalGet(key));
    return new Promise((resolve, reject) => {
      var _a;
      (_a = this.methods) == null ? void 0 : _a.get(this.getStorageKey(key)).then((res) => {
        const _res = isJsonString(res) ? JSON.parse(res) : res;
        logCodeManager.log("13:01:00", {
          tags: ["storage"],
          replacers: [this.key, this.getStorageKey(key)]
        }, _res);
        resolve(_res);
      }).catch(reject);
    });
  }
  set(key, value) {
    if (!this.ready || !this.methods.set)
      return;
    logCodeManager.log("13:01:01", {
      tags: ["storage"],
      replacers: [this.key, this.getStorageKey(key)]
    });
    this.methods.set(this.getStorageKey(key), JSON.stringify(value));
  }
  remove(key) {
    if (!this.ready || !this.methods.remove)
      return;
    logCodeManager.log("13:01:02", {
      tags: ["storage"],
      replacers: [this.key, this.getStorageKey(key)]
    });
    this.methods.remove(this.getStorageKey(key));
  }
  getStorageKey(key) {
    return this.config.prefix ? `_${this.config.prefix}_${key}` : key.toString();
  }
}

export { Storage };
