'use strict';

Object.defineProperty(exports, '__esModule', { value: true });

var utils = require('@agile-ts/utils');

function ownKeys(object, enumerableOnly) {
  var keys = Object.keys(object);

  if (Object.getOwnPropertySymbols) {
    var symbols = Object.getOwnPropertySymbols(object);

    if (enumerableOnly) {
      symbols = symbols.filter(function (sym) {
        return Object.getOwnPropertyDescriptor(object, sym).enumerable;
      });
    }

    keys.push.apply(keys, symbols);
  }

  return keys;
}

function _objectSpread2(target) {
  for (var i = 1; i < arguments.length; i++) {
    var source = arguments[i] != null ? arguments[i] : {};

    if (i % 2) {
      ownKeys(Object(source), true).forEach(function (key) {
        _defineProperty(target, key, source[key]);
      });
    } else if (Object.getOwnPropertyDescriptors) {
      Object.defineProperties(target, Object.getOwnPropertyDescriptors(source));
    } else {
      ownKeys(Object(source)).forEach(function (key) {
        Object.defineProperty(target, key, Object.getOwnPropertyDescriptor(source, key));
      });
    }
  }

  return target;
}

function asyncGeneratorStep(gen, resolve, reject, _next, _throw, key, arg) {
  try {
    var info = gen[key](arg);
    var value = info.value;
  } catch (error) {
    reject(error);
    return;
  }

  if (info.done) {
    resolve(value);
  } else {
    Promise.resolve(value).then(_next, _throw);
  }
}

function _asyncToGenerator(fn) {
  return function () {
    var self = this,
        args = arguments;
    return new Promise(function (resolve, reject) {
      var gen = fn.apply(self, args);

      function _next(value) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "next", value);
      }

      function _throw(err) {
        asyncGeneratorStep(gen, resolve, reject, _next, _throw, "throw", err);
      }

      _next(undefined);
    });
  };
}

function _defineProperty(obj, key, value) {
  if (key in obj) {
    Object.defineProperty(obj, key, {
      value: value,
      enumerable: true,
      configurable: true,
      writable: true
    });
  } else {
    obj[key] = value;
  }

  return obj;
}

var logCodeTypes = {
  '00': 'success',
  '01': 'info',
  '02': 'warn',
  '03': 'error'
};
var logCodeMessages = {
  '10:00:00': 'Created new AgileInstance.',
  '10:02:00': 'Be careful when binding multiple Agile Instances globally in one application!',
  '11:02:00': "The 'Local Storage' is not available in your current environment. " + "To use the '.persist()' functionality, please provide a custom Storage!",
  '11:02:01': 'The first allocated Storage for AgileTs must be set as the default Storage!',
  '11:03:00': "Storage with the key/name '${0}' already exists!",
  '11:03:01': "Couldn't find Storage '${0}'. " + "The Storage with the key/name '${0}' doesn't exists!",
  '11:03:02': "Storage with the key/name '${0}' isn't ready yet!",
  '11:03:03': 'No Storage found to get a value from! Please specify at least one Storage.',
  '11:03:04': 'No Storage found to store a value in! Please specify at least one Storage.',
  '11:03:05': 'No Storage found to remove a value from! Please specify at least one Storage.',
  '12:03:00': 'No valid persist Key found! Provide a valid key or assign one to the parent instance.',
  '12:03:01': 'No valid persist Storage Key found! Please specify at least one Storage Key to use the persist functionality.',
  '12:03:02': "Couldn't validate Persistent '${0}'." + "The Storage with the key/name '${1}' doesn't exists!`",
  '13:00:00': "Registered new Storage '${0}'.",
  '13:01:00': "GET value at key '${1}' from Storage '${0}'.",
  '13:01:01': "SET value at key '${1}' in Storage '${0}'.",
  '13:01:02': "REMOVE value at key '${1}' from Storage '${0}'.",
  '13:02:00': 'Using normalGet() in a async-based Storage might result in an unexpected return value. ' + 'Instead of a resolved value a Promise is returned!',
  '13:03:00': "Invalid Storage '${0}()' method provided!",
  '14:03:01': "'${1}' is a not supported type! Supported types are: String, Boolean, Array, Object, Number.",
  '14:03:02': "The 'patch()' method works only in object based States!",
  '14:03:03': 'Only one Interval can be active at once!',
  '14:03:04': "Failed to invert value of the type '${0}'!",
  '15:01:00': "Unregistered 'Callback' based Subscription.",
  '15:01:01': "Unregistered 'Component' based Subscription.",
  '15:01:02': "Registered 'Component' based Subscription.",
  '15:01:03': "Registered 'Callback' based Subscription.",
  '16:01:00': "Created Job '${0}'",
  '16:01:01': "Completed Job '${0}'",
  '16:01:02': 'Updated/Rerendered Subscriptions',
  '16:02:00': "SubscriptionContainer/Component '${0}' isn't ready to rerender!",
  '16:02:01': 'Job with not ready SubscriptionContainer/Component was removed from the runtime ' + 'after ${0} tries to avoid a Job overflow.',
  '17:03:00': "The 'perform()' method isn't set in Observer but need to be set! Observer is no stand alone class.",
  '18:00:00': "Integrated '${0}' into AgileTs '${1}'",
  '18:02:00': "Can't call the 'update()' method on a not ready Integration '${0}'!",
  '18:03:00': "Failed to integrate Framework '${0}' into AgileTs '${1}'!",
  '1A:02:00': 'Failed to build unique Item StorageKey!',
  '1A:02:01': 'Failed to build unique Group StorageKey!',
  '1B:02:00': "We recommend using 'createGroup()' " + "instead of 'Group()' outside the Collection configuration object.",
  '1B:02:01': "We recommend using 'createSelector()' " + "instead of 'Selector()' outside the Collection configuration object.",
  '1B:02:02': 'By overwriting the whole Item ' + "you have to pass the correct itemKey into the 'changes object!'",
  '1B:02:03': "We recommend using 'Group()' instead of 'createGroup()' " + 'inside the Collection configuration object.',
  '1B:02:04': "We recommend using 'Selector()' instead of 'createSelector()' " + 'inside the Collection configuration object.',
  '1B:02:05': "Collection '${0}' Item Data has to contain a primaryKey property called '${1}'!",
  '1B:03:00': "Couldn't update Item with the key/name '${0}' " + "because it doesn't exist in Collection '${1}'",
  '1B:03:01': "Valid object required to update Item value '${0}' in Collection '${1}'!",
  '1B:03:02': "Group with the key/name '${0}' already exists!",
  '1B:03:03': "Selector with the key/name '${0}' already exists!",
  '1B:03:04': "Couldn't update ItemKey from '${0}' to '${1}' " + "because an Item with the key/name '${1}' already exists in the Collection '${2}'!",
  '1B:03:05': "Item Data of Collection '${0}' has to be a valid object!",
  '1B:03:06': "Item tried to add to the Collection '${0}' belongs to another Collection '${1}'!",
  '1C:02:00': "Couldn't find some Items in the Collection '${0}' " + "during the rebuild of the Group '${1}' output.",
  '1C:03:00': "The 'output' property of the Group '${0}' is a automatically generated readonly property " + 'that can only be mutated by the Group itself!',
  '1C:03:01': "The 'item' property of the Group '${0}' is a automatically generated readonly property " + 'that can only be mutated by the Group itself!',
  '20:03:00': 'Failed to get Agile Instance from',
  '20:03:01': "Failed to create global Instance at '${0}'",
  '20:03:02': "Required module '${0}' couldn't be retrieved!",
  '00:03:00': "The '${0}()' method isn't set in ${1} but need to be set!" + ' ${1} is no stand alone class.',
  '00:03:01': "'${0}' has to be of the type ${1}!"
};
class LogCodeManager {
  constructor(logCodeMessages, loggerPackage) {
    _defineProperty(this, "logCodeMessages", void 0);

    _defineProperty(this, "_loggerPackage", void 0);

    _defineProperty(this, "allowLogging", true);

    this.logCodeMessages = logCodeMessages;
    this._loggerPackage = loggerPackage;
  }

  get logger() {
    var _this$_loggerPackage;

    return ((_this$_loggerPackage = this._loggerPackage) === null || _this$_loggerPackage === void 0 ? void 0 : _this$_loggerPackage.sharedLogger) || null;
  }

  getLog(logCode) {
    var replacers = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var result = this.logCodeMessages[logCode];
    if (result == null) return logCode;

    for (var i = 0; i < replacers.length; i++) {
      result = result.replace('${' + i + '}', replacers[i]);
    }

    return result;
  }

  log(logCode) {
    var _config$tags;

    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    if (this.logger != null && !this.logger.isActive || !this.allowLogging) return;
    config = utils.defineConfig(config, {
      replacers: [],
      tags: []
    });
    var logType = logCodeTypes[logCode.substr(3, 2)];
    if (typeof logType !== 'string') return;

    if (this.logger == null) {
      if (logType === 'error' || logType === 'warn') console[logType]("Agile: ".concat(this.getLog(logCode, config.replacers)));
      return;
    }

    var log = this.getLog(logCode, config.replacers);

    for (var _len = arguments.length, data = new Array(_len > 2 ? _len - 2 : 0), _key = 2; _key < _len; _key++) {
      data[_key - 2] = arguments[_key];
    }

    if (((_config$tags = config.tags) === null || _config$tags === void 0 ? void 0 : _config$tags.length) === 0) this.logger[logType](log, ...data);else this.logger.if.tag(config.tags)[logType](log, ...data);
  }

}
function assignAdditionalLogs(logCodeManager, additionalLogs) {
  var copiedLogCodeManager = new LogCodeManager(_objectSpread2(_objectSpread2({}, logCodeManager.logCodeMessages), additionalLogs), logCodeManager._loggerPackage);
  return copiedLogCodeManager;
}
var tempLogCodeManager;

if (process.env.NODE_ENV !== 'production') {
  var loggerPackage = null;

  try {
    loggerPackage = require('@agile-ts/logger');
  } catch (e) {}

  tempLogCodeManager = new LogCodeManager(logCodeMessages, loggerPackage);
} else {
  tempLogCodeManager = new LogCodeManager({}, null);
}

var logCodeManager = tempLogCodeManager;

class SubscriptionContainer {
  constructor(subs) {
    var _config;

    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _defineProperty(this, "key", void 0);

    _defineProperty(this, "ready", false);

    _defineProperty(this, "componentId", void 0);

    _defineProperty(this, "subscribers", void 0);

    _defineProperty(this, "updatedSubscribers", new Set());

    _defineProperty(this, "isObjectBased", false);

    _defineProperty(this, "subscriberKeysWeakMap", void 0);

    _defineProperty(this, "selectorsWeakMap", void 0);

    config = utils.defineConfig(config, {
      key: utils.generateId()
    });
    this.subscribers = new Set();
    this.key = config.key;
    this.componentId = (_config = config) === null || _config === void 0 ? void 0 : _config.componentId;
    this.subscriberKeysWeakMap = new WeakMap();
    this.selectorsWeakMap = new WeakMap();
    this.isObjectBased = !Array.isArray(subs);

    for (var key in subs) {
      var _config$proxyWeakMap, _config$proxyWeakMap$, _config$selectorWeakM, _config$selectorWeakM2;

      this.addSubscription(subs[key], {
        proxyPaths: (_config$proxyWeakMap = config.proxyWeakMap) === null || _config$proxyWeakMap === void 0 ? void 0 : (_config$proxyWeakMap$ = _config$proxyWeakMap.get(subs[key])) === null || _config$proxyWeakMap$ === void 0 ? void 0 : _config$proxyWeakMap$.paths,
        selectorMethods: (_config$selectorWeakM = config.selectorWeakMap) === null || _config$selectorWeakM === void 0 ? void 0 : (_config$selectorWeakM2 = _config$selectorWeakM.get(subs[key])) === null || _config$selectorWeakM2 === void 0 ? void 0 : _config$selectorWeakM2.methods,
        key: this.isObjectBased ? key : undefined
      });
    }
  }

  addSubscription(sub) {
    var _config$selectorMetho, _config$proxyPaths, _this$selectorsWeakMa, _this$selectorsWeakMa2;

    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    var toAddSelectorMethods = (_config$selectorMetho = config.selectorMethods) !== null && _config$selectorMetho !== void 0 ? _config$selectorMetho : [];
    var proxyPaths = (_config$proxyPaths = config.proxyPaths) !== null && _config$proxyPaths !== void 0 ? _config$proxyPaths : [];

    var _loop = function _loop(path) {
      toAddSelectorMethods.push(value => {
        var _value = value;

        for (var branch of path) {
          if (!utils.isValidObject(_value, true)) break;
          _value = _value[branch];
        }

        return _value;
      });
    };

    for (var path of proxyPaths) {
      _loop(path);
    }

    var existingSelectorMethods = (_this$selectorsWeakMa = (_this$selectorsWeakMa2 = this.selectorsWeakMap.get(sub)) === null || _this$selectorsWeakMa2 === void 0 ? void 0 : _this$selectorsWeakMa2.methods) !== null && _this$selectorsWeakMa !== void 0 ? _this$selectorsWeakMa : [];
    var newSelectorMethods = existingSelectorMethods.concat(toAddSelectorMethods);
    if (newSelectorMethods.length > 0) this.selectorsWeakMap.set(sub, {
      methods: newSelectorMethods
    });
    if (config.key != null) this.subscriberKeysWeakMap.set(sub, config.key);
    this.subscribers.add(sub);
    sub.subscribedTo.add(this);
  }

  removeSubscription(sub) {
    if (this.subscribers.has(sub)) {
      this.selectorsWeakMap.delete(sub);
      this.subscriberKeysWeakMap.delete(sub);
      this.subscribers.delete(sub);
      sub.subscribedTo.delete(this);
    }
  }

}

class CallbackSubscriptionContainer extends SubscriptionContainer {
  constructor(callback, subs) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    super(subs, config);

    _defineProperty(this, "callback", void 0);

    this.callback = callback;
  }

}

class ComponentSubscriptionContainer extends SubscriptionContainer {
  constructor(component, subs) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    super(subs, config);

    _defineProperty(this, "component", void 0);

    this.component = component;
  }

}

class SubController {
  constructor(agileInstance) {
    _defineProperty(this, "agileInstance", void 0);

    _defineProperty(this, "componentSubs", new Set());

    _defineProperty(this, "callbackSubs", new Set());

    _defineProperty(this, "mountedComponents", new Set());

    this.agileInstance = () => agileInstance;
  }

  subscribe(integrationInstance, subs) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    config = utils.defineConfig(config, {
      waitForMount: this.agileInstance().config.waitForMount
    });
    var subscriptionContainer = this[utils.isFunction(integrationInstance) ? 'createCallbackSubscriptionContainer' : 'createComponentSubscriptionContainer'](integrationInstance, subs, config);

    if (subscriptionContainer.isObjectBased && !Array.isArray(subs)) {
      var props = {};

      for (var key in subs) {
        if (subs[key]['value']) props[key] = subs[key]['value'];
      }

      return {
        subscriptionContainer,
        props
      };
    }

    return subscriptionContainer;
  }

  unsubscribe(subscriptionInstance) {
    var unsub = subscriptionContainer => {
      subscriptionContainer.ready = false;
      subscriptionContainer.subscribers.forEach(observer => {
        subscriptionContainer.removeSubscription(observer);
      });
    };

    if (subscriptionInstance instanceof CallbackSubscriptionContainer) {
      unsub(subscriptionInstance);
      this.callbackSubs.delete(subscriptionInstance);

      if (process.env.NODE_ENV !== 'production') {
        logCodeManager.log('15:01:00', {
          tags: ['subscription']
        }, subscriptionInstance);
      }

      return;
    }

    if (subscriptionInstance instanceof ComponentSubscriptionContainer) {
      unsub(subscriptionInstance);
      this.componentSubs.delete(subscriptionInstance);

      if (process.env.NODE_ENV !== 'production') {
        logCodeManager.log('15:01:01', {
          tags: ['subscription']
        }, subscriptionInstance);
      }

      return;
    }

    if (subscriptionInstance['componentSubscriptionContainers'] !== null && Array.isArray(subscriptionInstance.componentSubscriptionContainers)) {
      subscriptionInstance.componentSubscriptionContainers.forEach(subContainer => {
        unsub(subContainer);
        this.componentSubs.delete(subContainer);

        if (process.env.NODE_ENV !== 'production') {
          logCodeManager.log('15:01:01', {
            tags: ['subscription']
          }, subscriptionInstance);
        }
      });
      return;
    }
  }

  createComponentSubscriptionContainer(componentInstance, subs) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    var componentSubscriptionContainer = new ComponentSubscriptionContainer(componentInstance, subs, config);
    this.componentSubs.add(componentSubscriptionContainer);

    if (config.waitForMount) {
      if (this.mountedComponents.has(componentInstance)) componentSubscriptionContainer.ready = true;
    } else componentSubscriptionContainer.ready = true;

    if (Array.isArray(componentInstance.componentSubscriptionContainers)) componentInstance.componentSubscriptionContainers.push(componentSubscriptionContainer);else componentInstance['componentSubscriptionContainers'] = [componentSubscriptionContainer];

    if (process.env.NODE_ENV !== 'production') {
      logCodeManager.log('15:01:02', {
        tags: ['subscription']
      }, componentSubscriptionContainer);
    }

    return componentSubscriptionContainer;
  }

  createCallbackSubscriptionContainer(callbackFunction, subs) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    var callbackSubscriptionContainer = new CallbackSubscriptionContainer(callbackFunction, subs, config);
    this.callbackSubs.add(callbackSubscriptionContainer);
    callbackSubscriptionContainer.ready = true;

    if (process.env.NODE_ENV !== 'production') {
      logCodeManager.log('15:01:03', {
        tags: ['subscription']
      }, callbackSubscriptionContainer);
    }

    return callbackSubscriptionContainer;
  }

  mount(componentInstance) {
    if (componentInstance['componentSubscriptionContainers'] != null && Array.isArray(componentInstance.componentSubscriptionContainers)) componentInstance.componentSubscriptionContainers.map(c => c.ready = true);
    this.mountedComponents.add(componentInstance);
  }

  unmount(componentInstance) {
    if (componentInstance['componentSubscriptionContainers'] != null && Array.isArray(componentInstance.componentSubscriptionContainers)) componentInstance.componentSubscriptionContainers.map(c => c.ready = false);
    this.mountedComponents.delete(componentInstance);
  }

}

class Runtime {
  constructor(agileInstance) {
    _defineProperty(this, "agileInstance", void 0);

    _defineProperty(this, "currentJob", null);

    _defineProperty(this, "jobQueue", []);

    _defineProperty(this, "jobsToRerender", []);

    _defineProperty(this, "notReadyJobsToRerender", new Set());

    _defineProperty(this, "isPerformingJobs", false);

    _defineProperty(this, "bucketTimeout", null);

    this.agileInstance = () => agileInstance;
  }

  ingest(job) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      perform: !this.isPerformingJobs
    });
    this.jobQueue.push(job);

    if (process.env.NODE_ENV !== 'production') {
      logCodeManager.log('16:01:00', {
        tags: ['runtime'],
        replacers: [job.key]
      }, job);
    }

    if (config.perform) {
      var performJob = this.jobQueue.shift();
      if (performJob) this.perform(performJob);
    }
  }

  perform(job) {
    this.isPerformingJobs = true;
    this.currentJob = job;
    job.observer.perform(job);
    job.performed = true;
    job.observer.dependents.forEach(observer => observer.ingest());
    if (job.rerender) this.jobsToRerender.push(job);
    this.currentJob = null;

    if (process.env.NODE_ENV !== 'production') {
      logCodeManager.log('16:01:01', {
        tags: ['runtime'],
        replacers: [job.key]
      }, job);
    }

    if (this.jobQueue.length > 0) {
      var performJob = this.jobQueue.shift();
      if (performJob) this.perform(performJob);
    } else {
      this.isPerformingJobs = false;

      if (this.jobsToRerender.length > 0) {
        if (this.agileInstance().config.bucket) {
          if (this.bucketTimeout == null) {
            this.bucketTimeout = setTimeout(() => {
              this.bucketTimeout = null;
              this.updateSubscribers();
            });
          }
        } else this.updateSubscribers();
      }
    }
  }

  updateSubscribers() {
    var jobsToRerender = this.jobsToRerender.concat(Array.from(this.notReadyJobsToRerender));
    this.notReadyJobsToRerender = new Set();
    this.jobsToRerender = [];
    if (!this.agileInstance().hasIntegration() || jobsToRerender.length <= 0) return false;
    var subscriptionContainerToUpdate = this.extractToUpdateSubscriptionContainer(jobsToRerender);
    if (subscriptionContainerToUpdate.length <= 0) return false;
    this.updateSubscriptionContainer(subscriptionContainerToUpdate);
    return true;
  }

  extractToUpdateSubscriptionContainer(jobs) {
    var _this = this;

    var subscriptionsToUpdate = new Set();

    var _loop = function _loop(i) {
      var job = jobs[i];
      job.subscriptionContainersToUpdate.forEach(subscriptionContainer => {
        var updateSubscriptionContainer = true;

        if (!subscriptionContainer.ready) {
          if (!job.config.maxTriesToUpdate || job.timesTriedToUpdateCount < job.config.maxTriesToUpdate) {
            job.timesTriedToUpdateCount++;

            _this.notReadyJobsToRerender.add(job);

            if (process.env.NODE_ENV !== 'production') {
              logCodeManager.log('16:02:00', {
                replacers: [subscriptionContainer.key]
              }, subscriptionContainer);
            }
          } else {
            if (process.env.NODE_ENV !== 'production') {
              logCodeManager.log('16:02:01', {
                replacers: [job.config.maxTriesToUpdate]
              }, subscriptionContainer);
            }
          }

          return;
        }

        updateSubscriptionContainer = updateSubscriptionContainer && _this.handleSelectors(subscriptionContainer, job);

        if (updateSubscriptionContainer) {
          subscriptionContainer.updatedSubscribers.add(job.observer);
          subscriptionsToUpdate.add(subscriptionContainer);
        }

        job.subscriptionContainersToUpdate.delete(subscriptionContainer);
      });
    };

    for (var i = 0; i < jobs.length; i++) {
      _loop(i);
    }

    return Array.from(subscriptionsToUpdate);
  }

  updateSubscriptionContainer(subscriptionsToUpdate) {
    for (var i = 0; i < subscriptionsToUpdate.length; i++) {
      var subscriptionContainer = subscriptionsToUpdate[i];
      if (subscriptionContainer instanceof CallbackSubscriptionContainer) subscriptionContainer.callback();
      if (subscriptionContainer instanceof ComponentSubscriptionContainer) this.agileInstance().integrations.update(subscriptionContainer.component, this.getUpdatedObserverValues(subscriptionContainer));
      subscriptionContainer.updatedSubscribers.clear();
    }

    if (process.env.NODE_ENV !== 'production') {
      logCodeManager.log('16:01:02', {
        tags: ['runtime']
      }, subscriptionsToUpdate);
    }
  }

  getUpdatedObserverValues(subscriptionContainer) {
    var props = {};

    for (var observer of subscriptionContainer.updatedSubscribers) {
      var _subscriptionContaine;

      var key = (_subscriptionContaine = subscriptionContainer.subscriberKeysWeakMap.get(observer)) !== null && _subscriptionContaine !== void 0 ? _subscriptionContaine : observer.key;
      if (key != null) props[key] = observer.value;
    }

    return props;
  }

  handleSelectors(subscriptionContainer, job) {
    var _subscriptionContaine2;

    var selectorMethods = (_subscriptionContaine2 = subscriptionContainer.selectorsWeakMap.get(job.observer)) === null || _subscriptionContaine2 === void 0 ? void 0 : _subscriptionContaine2.methods;
    if (selectorMethods == null) return true;
    var previousValue = job.observer.previousValue;
    var newValue = job.observer.value;
    if (!previousValue && !newValue) return false;

    for (var selectorMethod of selectorMethods) {
      if (utils.notEqual(selectorMethod(newValue), selectorMethod(previousValue))) return true;
    }

    return false;
  }

}

class RuntimeJob {
  constructor(observer) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _defineProperty(this, "config", void 0);

    _defineProperty(this, "key", void 0);

    _defineProperty(this, "observer", void 0);

    _defineProperty(this, "rerender", void 0);

    _defineProperty(this, "subscriptionContainersToUpdate", new Set());

    _defineProperty(this, "timesTriedToUpdateCount", 0);

    _defineProperty(this, "performed", false);

    config = utils.defineConfig(config, {
      background: false,
      sideEffects: {
        enabled: true,
        exclude: []
      },
      force: false,
      maxTriesToUpdate: 3,
      any: {}
    });
    this.config = {
      background: config.background,
      force: config.force,
      sideEffects: config.sideEffects,
      maxTriesToUpdate: config.maxTriesToUpdate,
      any: config.any
    };
    this.observer = observer;
    this.rerender = !config.background && this.observer.agileInstance().integrations.hasIntegration();
    this.key = config.key;
    this.subscriptionContainersToUpdate = new Set(observer.subscribedTo);
  }

}

class Observer {
  constructor(agileInstance) {
    var _config$dependents, _config$subs;

    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _defineProperty(this, "agileInstance", void 0);

    _defineProperty(this, "key", void 0);

    _defineProperty(this, "dependents", new Set());

    _defineProperty(this, "subscribedTo", new Set());

    _defineProperty(this, "value", void 0);

    _defineProperty(this, "previousValue", void 0);

    config = utils.defineConfig(config, {
      dependents: [],
      subs: [],
      value: null
    });

    this.agileInstance = () => agileInstance;

    this.key = config.key;
    this.value = config.value;
    this.previousValue = config.value;
    (_config$dependents = config.dependents) === null || _config$dependents === void 0 ? void 0 : _config$dependents.forEach(observer => this.addDependent(observer));
    (_config$subs = config.subs) === null || _config$subs === void 0 ? void 0 : _config$subs.forEach(subscriptionContainer => subscriptionContainer.addSubscription(this));
  }

  ingest() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    config = utils.defineConfig(config, {
      perform: true,
      key: logCodeManager.allowLogging ? "".concat(this.key != null ? this.key + '_' : '').concat(utils.generateId()) : undefined
    });
    var job = new RuntimeJob(this, config);
    this.agileInstance().runtime.ingest(job, {
      perform: config.perform
    });
  }

  perform(job) {
    if (process.env.NODE_ENV !== 'production') {
      logCodeManager.log('17:03:00');
    }
  }

  addDependent(observer) {
    if (!this.dependents.has(observer)) this.dependents.add(observer);
  }

  removeDependent(observer) {
    if (this.dependents.has(observer)) this.dependents.delete(observer);
  }

}

function extractObservers(instances) {
  var observers = [];
  var tempInstancesArray = utils.normalizeArray(instances, true);

  for (var instance of tempInstancesArray) {
    if (instance == null) {
      observers.push({});
      continue;
    }

    if (instance.isCollection) {
      observers.push(instance.getGroupWithReference(instance.config.defaultGroupKey).observers);
      continue;
    }

    if (instance['observer'] && instance['observer'] instanceof Observer) {
      observers.push({
        value: instance['observer']
      });
      continue;
    }

    if (instance['observers']) {
      var extractedObservers = {};

      for (var key in instance['observers']) {
        if (instance['observers'][key] instanceof Observer) {
          extractedObservers[key] = instance['observers'][key];
        }
      }

      observers.push(extractedObservers);
      continue;
    }

    if (instance instanceof Observer) {
      observers.push({
        value: instance
      });
      continue;
    }

    observers.push({});
  }

  return Array.isArray(instances) ? observers : observers[0];
}
function extractRelevantObservers(instances, observerType) {
  var depsWithIndicator = {};
  var depsWithNoIndicator = [];

  for (var depKey in instances) {
    var _extractedObservers$o;

    var extractedObservers = extractObservers(instances[depKey]);
    var observer = undefined;
    if (observerType != null && extractedObservers[observerType] != null) observer = extractedObservers[observerType];
    if (observerType == null) observer = (_extractedObservers$o = extractedObservers['output']) !== null && _extractedObservers$o !== void 0 ? _extractedObservers$o : extractedObservers['value'];
    if (Array.isArray(instances)) depsWithNoIndicator.push(observer);else depsWithIndicator[depKey] = observer;
  }

  return Array.isArray(instances) ? depsWithNoIndicator : depsWithIndicator;
}
function optionalRequire(moduleName) {
  var error = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;
  var requiredPackage = null;

  try {
    requiredPackage = require(moduleName);
  } catch (e) {
    if (error) {
      logCodeManager.log('20:03:02', {
        replacers: [moduleName]
      });
      console.log(e);
    }
  }

  return requiredPackage;
}
function globalBind(key, instance) {
  var overwrite = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : false;

  try {
    if (overwrite) {
      globalThis[key] = instance;
      return true;
    }

    if (globalThis[key] == null) {
      globalThis[key] = instance;
      return true;
    }
  } catch (e) {
    logCodeManager.log('20:03:01', {
      replacers: [key]
    });
  }

  return false;
}
var runsOnServer = () => {
  return !(typeof window !== 'undefined' && typeof window.document !== 'undefined' && typeof window.document.createElement !== 'undefined');
};

class Integration {
  constructor(config) {
    _defineProperty(this, "key", void 0);

    _defineProperty(this, "frameworkInstance", void 0);

    _defineProperty(this, "ready", false);

    _defineProperty(this, "integrated", false);

    _defineProperty(this, "methods", void 0);

    this.key = config.key;
    this.frameworkInstance = config.frameworkInstance;
    this.methods = {
      bind: config.bind,
      updateMethod: config.updateMethod
    };
  }

}

var onRegisterInitialIntegrationCallbacks = [];
class Integrations {
  static addInitialIntegration(integration) {
    if (integration instanceof Integration) {
      onRegisterInitialIntegrationCallbacks.forEach(callback => callback(integration));
      Integrations.initialIntegrations.push(integration);
    }
  }

  static onRegisterInitialIntegration(callback) {
    onRegisterInitialIntegrationCallbacks.push(callback);
    Integrations.initialIntegrations.forEach(integration => {
      callback(integration);
    });
  }

  constructor(agileInstance) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _defineProperty(this, "agileInstance", void 0);

    _defineProperty(this, "integrations", new Set());

    config = utils.defineConfig(config, {
      autoIntegrate: true
    });

    this.agileInstance = () => agileInstance;

    if (config.autoIntegrate) {
      Integrations.onRegisterInitialIntegration(integration => {
        this.integrate(integration);
      });
    }
  }

  integrate(integration) {
    var _this = this;

    return _asyncToGenerator(function* () {
      if (integration.key == null) {
        if (process.env.NODE_ENV !== 'production') {
          logCodeManager.log('18:03:00', {
            replacers: [integration.key, _this.agileInstance().key]
          }, integration);
        }

        return false;
      }

      if (integration.methods.bind) integration.ready = yield integration.methods.bind(_this.agileInstance());else integration.ready = true;

      _this.integrations.add(integration);

      integration.integrated = true;

      if (process.env.NODE_ENV !== 'production') {
        logCodeManager.log('18:00:00', {
          replacers: [integration.key, _this.agileInstance().key]
        }, integration);
      }

      return true;
    })();
  }

  update(componentInstance, updatedData) {
    this.integrations.forEach(integration => {
      if (!integration.ready) {
        logCodeManager.log('18:02:00', {
          replacers: [integration.key]
        });
        return;
      }

      if (integration.methods.updateMethod) integration.methods.updateMethod(componentInstance, updatedData);
    });
  }

  hasIntegration() {
    return this.integrations.size > 0;
  }

}

_defineProperty(Integrations, "initialIntegrations", []);

class Agile$1 {
  constructor() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    _defineProperty(this, "config", void 0);

    _defineProperty(this, "key", void 0);

    _defineProperty(this, "runtime", void 0);

    _defineProperty(this, "subController", void 0);

    _defineProperty(this, "integrations", void 0);

    config = utils.defineConfig(config, {
      localStorage: false,
      waitForMount: true,
      bindGlobal: false,
      autoIntegrate: true,
      bucket: true
    });
    this.config = {
      waitForMount: config.waitForMount,
      bucket: config.bucket
    };
    this.key = config.key;
    this.integrations = new Integrations(this, {
      autoIntegrate: config.autoIntegrate
    });
    this.runtime = new Runtime(this);
    this.subController = new SubController(this);
    logCodeManager.log('10:00:00', {}, this);
    if (config.bindGlobal) if (!globalBind(Agile$1.globalKey, this)) {
      logCodeManager.log('10:02:00');
    }
  }

  integrate(integration) {
    this.integrations.integrate(integration);
    return this;
  }

  hasIntegration() {
    return this.integrations.hasIntegration();
  }

}

_defineProperty(Agile$1, "globalKey", '__agile__');

class Storage {
  constructor(config) {
    _defineProperty(this, "config", void 0);

    _defineProperty(this, "key", void 0);

    _defineProperty(this, "ready", false);

    _defineProperty(this, "methods", void 0);

    config = utils.defineConfig(config, {
      prefix: 'agile',
      async: false
    });
    this.key = config.key;
    this.methods = config.methods;
    this.config = {
      prefix: config.prefix,
      async: config.async
    };
    this.ready = this.validate();
    if (!this.ready) return;
    if (utils.isAsyncFunction(this.methods.get) || utils.isAsyncFunction(this.methods.set) || utils.isAsyncFunction(this.methods.remove)) this.config.async = true;
  }

  validate() {
    var _this$methods, _this$methods2, _this$methods3;

    if (!utils.isFunction((_this$methods = this.methods) === null || _this$methods === void 0 ? void 0 : _this$methods.get)) {
      logCodeManager.log('13:03:00', {
        replacers: ['get']
      });
      return false;
    }

    if (!utils.isFunction((_this$methods2 = this.methods) === null || _this$methods2 === void 0 ? void 0 : _this$methods2.set)) {
      logCodeManager.log('13:03:00', {
        replacers: ['set']
      });
      return false;
    }

    if (!utils.isFunction((_this$methods3 = this.methods) === null || _this$methods3 === void 0 ? void 0 : _this$methods3.remove)) {
      logCodeManager.log('13:03:00', {
        replacers: ['remove']
      });
      return false;
    }

    return true;
  }

  normalGet(key) {
    if (!this.ready || !this.methods.get) return undefined;
    if (utils.isAsyncFunction(this.methods.get)) logCodeManager.log('13:02:00');
    var res = this.methods.get(this.getStorageKey(key));

    var _res = utils.isJsonString(res) ? JSON.parse(res) : res;

    logCodeManager.log('13:01:00', {
      tags: ['storage'],
      replacers: [this.key, this.getStorageKey(key)]
    }, _res);
    return _res;
  }

  get(key) {
    if (!this.ready || !this.methods.get) return Promise.resolve(undefined);
    if (!utils.isAsyncFunction(this.methods.get)) return Promise.resolve(this.normalGet(key));
    return new Promise((resolve, reject) => {
      var _this$methods4;

      (_this$methods4 = this.methods) === null || _this$methods4 === void 0 ? void 0 : _this$methods4.get(this.getStorageKey(key)).then(res => {
        var _res = utils.isJsonString(res) ? JSON.parse(res) : res;

        logCodeManager.log('13:01:00', {
          tags: ['storage'],
          replacers: [this.key, this.getStorageKey(key)]
        }, _res);
        resolve(_res);
      }).catch(reject);
    });
  }

  set(key, value) {
    if (!this.ready || !this.methods.set) return;
    logCodeManager.log('13:01:01', {
      tags: ['storage'],
      replacers: [this.key, this.getStorageKey(key)]
    });
    this.methods.set(this.getStorageKey(key), JSON.stringify(value));
  }

  remove(key) {
    if (!this.ready || !this.methods.remove) return;
    logCodeManager.log('13:01:02', {
      tags: ['storage'],
      replacers: [this.key, this.getStorageKey(key)]
    });
    this.methods.remove(this.getStorageKey(key));
  }

  getStorageKey(key) {
    return this.config.prefix ? "_".concat(this.config.prefix, "_").concat(key) : key.toString();
  }

}

class Storages {
  constructor(agileInstance) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _defineProperty(this, "agileInstance", void 0);

    _defineProperty(this, "config", void 0);

    _defineProperty(this, "storages", {});

    _defineProperty(this, "persistentInstances", {});

    this.agileInstance = () => agileInstance;

    config = utils.defineConfig(config, {
      localStorage: false,
      defaultStorageKey: null
    });
    this.config = {
      defaultStorageKey: config.defaultStorageKey
    };
    if (config.localStorage) this.instantiateLocalStorage();
  }

  instantiateLocalStorage() {
    if (!Storages.localStorageAvailable()) {
      logCodeManager.log('11:02:00');
      return false;
    }

    var _localStorage = new Storage({
      key: 'localStorage',
      async: false,
      methods: {
        get: localStorage.getItem.bind(localStorage),
        set: localStorage.setItem.bind(localStorage),
        remove: localStorage.removeItem.bind(localStorage)
      }
    });

    return this.register(_localStorage, {
      default: true
    });
  }

  register(storage) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    var hasRegisteredAnyStorage = utils.notEqual(this.storages, {});

    if (Object.prototype.hasOwnProperty.call(this.storages, storage.key)) {
      logCodeManager.log('11:03:00', {
        replacers: [storage.key]
      });
      return false;
    }

    if (!hasRegisteredAnyStorage && config.default === false) logCodeManager.log('11:02:01');
    if (!hasRegisteredAnyStorage) config.default = true;
    this.storages[storage.key] = storage;
    if (config.default) this.config.defaultStorageKey = storage.key;

    for (var persistentKey of Object.keys(this.persistentInstances)) {
      var persistent = this.persistentInstances[persistentKey];
      if (persistent == null) continue;

      if (persistent.storageKeys.includes(storage.key)) {
        var isValid = persistent.validatePersistent();
        if (isValid) persistent.initialLoading();
        continue;
      }

      if (persistent.config.defaultStorageKey == null) {
        persistent.assignStorageKeys();

        var _isValid = persistent.validatePersistent();

        if (_isValid) persistent.initialLoading();
      }
    }

    logCodeManager.log('13:00:00', {
      replacers: [storage.key]
    }, storage);
    return true;
  }

  getStorage(storageKey) {
    if (!storageKey) return undefined;
    var storage = this.storages[storageKey];

    if (!storage) {
      logCodeManager.log('11:03:01', {
        replacers: [storageKey]
      });
      return undefined;
    }

    if (!storage.ready) {
      logCodeManager.log('11:03:02', {
        replacers: [storageKey]
      });
      return undefined;
    }

    return storage;
  }

  get(storageItemKey, storageKey) {
    if (!this.hasStorage()) {
      logCodeManager.log('11:03:03');
      return Promise.resolve(undefined);
    }

    if (storageKey) {
      var storage = this.getStorage(storageKey);
      if (storage) return storage.get(storageItemKey);
    }

    var defaultStorage = this.getStorage(this.config.defaultStorageKey);
    return (defaultStorage === null || defaultStorage === void 0 ? void 0 : defaultStorage.get(storageItemKey)) || Promise.resolve(undefined);
  }

  set(storageItemKey, value, storageKeys) {
    if (!this.hasStorage()) {
      logCodeManager.log('11:03:04');
      return;
    }

    if (storageKeys != null) {
      for (var storageKey of storageKeys) {
        var _this$getStorage;

        (_this$getStorage = this.getStorage(storageKey)) === null || _this$getStorage === void 0 ? void 0 : _this$getStorage.set(storageItemKey, value);
      }

      return;
    }

    var defaultStorage = this.getStorage(this.config.defaultStorageKey);
    defaultStorage === null || defaultStorage === void 0 ? void 0 : defaultStorage.set(storageItemKey, value);
  }

  remove(storageItemKey, storageKeys) {
    if (!this.hasStorage()) {
      logCodeManager.log('11:03:05');
      return;
    }

    if (storageKeys) {
      for (var storageKey of storageKeys) {
        var _this$getStorage2;

        (_this$getStorage2 = this.getStorage(storageKey)) === null || _this$getStorage2 === void 0 ? void 0 : _this$getStorage2.remove(storageItemKey);
      }

      return;
    }

    var defaultStorage = this.getStorage(this.config.defaultStorageKey);
    defaultStorage === null || defaultStorage === void 0 ? void 0 : defaultStorage.remove(storageItemKey);
  }

  hasStorage() {
    return utils.notEqual(this.storages, {});
  }

  static localStorageAvailable() {
    try {
      localStorage.setItem('_myDummyKey_', 'myDummyValue');
      localStorage.removeItem('_myDummyKey_');
      return true;
    } catch (e) {
      return false;
    }
  }

}

exports.sharedAgileInstance = new Agile$1({
  key: 'shared'
});
function assignSharedAgileInstance(agileInstance) {
  exports.sharedAgileInstance = agileInstance;
}
function getAgileInstance(instance) {
  try {
    if (instance != null) {
      var _agileInstance = utils.isFunction(instance['agileInstance']) ? instance['agileInstance']() : instance['agileInstance'];

      if (_agileInstance) return _agileInstance;
    }

    if (exports.sharedAgileInstance instanceof Agile$1) {
      return exports.sharedAgileInstance;
    }

    return globalThis[Agile$1.globalKey];
  } catch (e) {
    logCodeManager.log('20:03:00', {}, instance);
  }

  return undefined;
}

exports.sharedStorageManager = null;
var assignSharedStorageManager = instance => {
  exports.sharedStorageManager = instance;
};
function createStorageManager() {
  var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
  config = utils.defineConfig(config, {
    agileInstance: exports.sharedAgileInstance
  });
  return new Storages(config.agileInstance, config);
}
function getSharedStorageManager() {
  if (exports.sharedStorageManager == null) {
    var newStorageManager = createStorageManager({
      localStorage: !runsOnServer()
    });
    assignSharedStorageManager(newStorageManager);
    return newStorageManager;
  }

  return exports.sharedStorageManager;
}

class Persistent {
  constructor(agileInstance) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _defineProperty(this, "agileInstance", void 0);

    _defineProperty(this, "config", void 0);

    _defineProperty(this, "_key", void 0);

    _defineProperty(this, "ready", false);

    _defineProperty(this, "isPersisted", false);

    _defineProperty(this, "onLoad", void 0);

    _defineProperty(this, "storageKeys", []);

    this.agileInstance = () => agileInstance;

    this._key = Persistent.placeHolderKey;
    config = utils.defineConfig(config, {
      loadValue: true,
      storageKeys: [],
      defaultStorageKey: null
    });
    this.config = {
      defaultStorageKey: config.defaultStorageKey
    };

    if (config.loadValue) {
      this.instantiatePersistent({
        storageKeys: config.storageKeys,
        key: config.key,
        defaultStorageKey: config.defaultStorageKey
      });
    }
  }

  set key(value) {
    this.setKey(value);
  }

  get key() {
    return this._key;
  }

  setKey(value) {
    var _this = this;

    return _asyncToGenerator(function* () {
      var oldKey = _this._key;
      var wasReady = _this.ready;
      if (value === _this._key) return;
      _this._key = value !== null && value !== void 0 ? value : Persistent.placeHolderKey;

      var isValid = _this.validatePersistent();

      if (!wasReady) {
        if (isValid) yield _this.initialLoading();
        return;
      }

      yield _this.removePersistedValue(oldKey);
      if (isValid) yield _this.persistValue(value);
    })();
  }

  instantiatePersistent() {
    var _this$formatKey;

    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    this._key = (_this$formatKey = this.formatKey(config.key)) !== null && _this$formatKey !== void 0 ? _this$formatKey : Persistent.placeHolderKey;
    this.assignStorageKeys(config.storageKeys, config.defaultStorageKey);
    this.validatePersistent();
    var storageManager = getSharedStorageManager();

    if (this._key !== Persistent.placeHolderKey && storageManager != null) {
      storageManager.persistentInstances[this._key] = this;
    }
  }

  validatePersistent() {
    var isValid = true;

    if (this._key === Persistent.placeHolderKey) {
      logCodeManager.log('12:03:00');
      isValid = false;
    }

    if (this.config.defaultStorageKey == null || this.storageKeys.length <= 0) {
      logCodeManager.log('12:03:01');
      isValid = false;
    }

    this.storageKeys.map(key => {
      var _getSharedStorageMana;

      if (!((_getSharedStorageMana = getSharedStorageManager()) !== null && _getSharedStorageMana !== void 0 && _getSharedStorageMana.storages[key])) {
        logCodeManager.log('12:03:02', {
          replacers: [this._key, key]
        });
        isValid = false;
      }
    });
    this.ready = isValid;
    return isValid;
  }

  assignStorageKeys() {
    var storageKeys = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    var defaultStorageKey = arguments.length > 1 ? arguments[1] : undefined;

    var _storageKeys = utils.copy(storageKeys);

    if (defaultStorageKey != null && !_storageKeys.includes(defaultStorageKey)) _storageKeys.push(defaultStorageKey);

    if (_storageKeys.length <= 0) {
      var _getSharedStorageMana2;

      var _defaultStorageKey = (_getSharedStorageMana2 = getSharedStorageManager()) === null || _getSharedStorageMana2 === void 0 ? void 0 : _getSharedStorageMana2.config.defaultStorageKey;

      if (_defaultStorageKey != null) {
        var _getSharedStorageMana3;

        this.config.defaultStorageKey = _defaultStorageKey;

        _storageKeys.push((_getSharedStorageMana3 = getSharedStorageManager()) === null || _getSharedStorageMana3 === void 0 ? void 0 : _getSharedStorageMana3.config.defaultStorageKey);
      }
    } else {
      this.config.defaultStorageKey = defaultStorageKey !== null && defaultStorageKey !== void 0 ? defaultStorageKey : _storageKeys[0];
    }

    this.storageKeys = _storageKeys;
  }

  initialLoading() {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      var success = yield _this2.loadPersistedValue();
      if (_this2.onLoad) _this2.onLoad(success);
      if (!success) yield _this2.persistValue();
    })();
  }

  loadPersistedValue(storageItemKey) {
    return _asyncToGenerator(function* () {
      if (process.env.NODE_ENV !== 'production') {
        logCodeManager.log('00:03:00', {
          replacers: ['loadPersistedValue', 'Persistent']
        });
      }

      return false;
    })();
  }

  persistValue(storageItemKey) {
    return _asyncToGenerator(function* () {
      if (process.env.NODE_ENV !== 'production') {
        logCodeManager.log('00:03:00', {
          replacers: ['persistValue', 'Persistent']
        });
      }

      return false;
    })();
  }

  removePersistedValue(storageItemKey) {
    return _asyncToGenerator(function* () {
      if (process.env.NODE_ENV !== 'production') {
        logCodeManager.log('00:03:00', {
          replacers: ['removePersistedValue', 'Persistent']
        });
      }

      return false;
    })();
  }

  formatKey(key) {
    return key;
  }

}

_defineProperty(Persistent, "placeHolderKey", '__THIS_IS_A_PLACEHOLDER__');

function createStorage(config) {
  return new Storage(config);
}

class StateRuntimeJob extends RuntimeJob {
  constructor(observer) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    super(observer, config);
    config = utils.defineConfig(config, {
      storage: true,
      overwrite: false
    });
    this.config['overwrite'] = config.overwrite;
    this.config['storage'] = config.storage;
  }

}

class StateObserver extends Observer {
  constructor(state) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    super(state.agileInstance(), utils.defineConfig(config, {
      value: utils.copy(state._value)
    }));

    _defineProperty(this, "state", void 0);

    _defineProperty(this, "nextStateValue", void 0);

    this.state = () => state;

    this.nextStateValue = utils.copy(state._value);
  }

  ingest() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var state = this.state();

    if (state.isComputed) {
      state.compute().then(result => {
        this.ingestValue(result, config);
      });
    } else {
      this.ingestValue(state.nextStateValue, config);
    }
  }

  ingestValue(newStateValue) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    var state = this.state();
    config = utils.defineConfig(config, {
      perform: true,
      force: false,
      key: logCodeManager.allowLogging ? "".concat(this.key != null ? this.key + '_' : '').concat(utils.generateId(), "_value") : undefined
    });

    if (state.isPlaceholder) {
      config.force = true;
      config.overwrite = true;
    }

    this.nextStateValue = state.computeValueMethod ? utils.copy(state.computeValueMethod(newStateValue)) : utils.copy(newStateValue);
    if (utils.equal(state._value, this.nextStateValue) && !config.force) return;
    var job = new StateRuntimeJob(this, config);
    this.agileInstance().runtime.ingest(job, {
      perform: config.perform
    });
  }

  perform(job) {
    var observer = job.observer;
    var state = observer.state();
    state.previousStateValue = utils.copy(state._value);
    state._value = utils.copy(observer.nextStateValue);
    state.nextStateValue = utils.copy(observer.nextStateValue);

    if (job.config.overwrite) {
      state.initialStateValue = utils.copy(state._value);
      state.previousStateValue = utils.copy(state._value);
      state.isPlaceholder = false;
    }

    state.isSet = utils.notEqual(state._value, state.initialStateValue);
    this.sideEffects(job);
    job.observer.previousValue = Object.freeze(utils.copy(observer.value));
    job.observer.value = utils.copy(state._value);
  }

  sideEffects(job) {
    var _job$config, _job$config$sideEffec;

    var state = job.observer.state();
    if (Object.keys(state.sideEffects).length === 0) return;

    if ((_job$config = job.config) !== null && _job$config !== void 0 && (_job$config$sideEffec = _job$config.sideEffects) !== null && _job$config$sideEffec !== void 0 && _job$config$sideEffec.enabled) {
      var sideEffectArray = utils.createArrayFromObject(state.sideEffects);
      sideEffectArray.sort(function (a, b) {
        return b.instance.weight - a.instance.weight;
      });

      for (var sideEffect of sideEffectArray) {
        if (utils.isFunction(sideEffect.instance.callback)) {
          var _job$config$sideEffec2;

          if (!((_job$config$sideEffec2 = job.config.sideEffects.exclude) !== null && _job$config$sideEffec2 !== void 0 && _job$config$sideEffec2.includes(sideEffect.key))) sideEffect.instance.callback(job.observer.state(), job.config);
        }
      }
    }
  }

}

class ComputedTracker {
  constructor() {}

  static track() {
    this.isTracking = true;
  }

  static tracked(observer) {
    if (this.isTracking) this.trackedObservers.add(observer);
  }

  static getTrackedObservers() {
    var trackedObservers = Array.from(this.trackedObservers);
    this.isTracking = false;
    this.trackedObservers = new Set();
    return trackedObservers;
  }

}

_defineProperty(ComputedTracker, "isTracking", false);

_defineProperty(ComputedTracker, "trackedObservers", new Set());

class State {
  constructor(agileInstance, initialValue) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    _defineProperty(this, "agileInstance", void 0);

    _defineProperty(this, "_key", void 0);

    _defineProperty(this, "isSet", false);

    _defineProperty(this, "isPlaceholder", false);

    _defineProperty(this, "initialStateValue", void 0);

    _defineProperty(this, "_value", void 0);

    _defineProperty(this, "previousStateValue", void 0);

    _defineProperty(this, "nextStateValue", void 0);

    _defineProperty(this, "observers", {});

    _defineProperty(this, "sideEffects", {});

    config = utils.defineConfig(config, {
      dependents: [],
      isPlaceholder: false
    });

    this.agileInstance = () => agileInstance;

    this._key = config.key;
    this.observers['value'] = new StateObserver(this, {
      key: config.key,
      dependents: config.dependents
    });
    this.initialStateValue = utils.copy(initialValue);
    this._value = utils.copy(initialValue);
    this.previousStateValue = utils.copy(initialValue);
    this.nextStateValue = utils.copy(initialValue);
    this.isPlaceholder = true;
    if (!config.isPlaceholder) this.set(initialValue, {
      overwrite: true
    });
  }

  set value(value) {
    this.set(value);
  }

  get value() {
    ComputedTracker.tracked(this.observers['value']);
    return utils.copy(this._value);
  }

  set key(value) {
    this.setKey(value);
  }

  get key() {
    return this._key;
  }

  setKey(value) {
    this._key = value;

    for (var observerKey in this.observers) {
      this.observers[observerKey].key = value;
    }

    return this;
  }

  set(value) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      force: false
    });

    var _value = utils.isFunction(value) ? value(utils.copy(this._value)) : value;

    this.observers['value'].ingestValue(_value, config);
    return this;
  }

  ingest() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    this.observers['value'].ingest(config);
    return this;
  }

  addSideEffect(key, callback) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    config = utils.defineConfig(config, {
      weight: 10
    });

    if (!utils.isFunction(callback)) {
      logCodeManager.log('00:03:01', {
        replacers: ['Side Effect Callback', 'function']
      });
      return this;
    }

    this.sideEffects[key] = {
      callback: callback,
      weight: config.weight
    };
    return this;
  }

  removeSideEffect(key) {
    delete this.sideEffects[key];
    return this;
  }

  hasSideEffect(key) {
    return !!this.sideEffects[key];
  }

}

class StatePersistent extends Persistent {
  constructor(state) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    super(state.agileInstance(), {
      loadValue: false
    });

    _defineProperty(this, "state", void 0);

    _defineProperty(this, "onMigrate", void 0);

    _defineProperty(this, "onSave", void 0);

    config = utils.defineConfig(config, {
      loadValue: true,
      storageKeys: [],
      defaultStorageKey: null
    });

    this.state = () => state;

    this.onMigrate = config.onMigrate;
    this.onSave = config.onSave;
    this.instantiatePersistent({
      key: config.key,
      storageKeys: config.storageKeys,
      defaultStorageKey: config.defaultStorageKey
    });
    if (this.ready && config.loadValue) this.initialLoading();
  }

  initialLoading() {
    var _superprop_getInitialLoading = () => super.initialLoading,
        _this = this;

    return _asyncToGenerator(function* () {
      _superprop_getInitialLoading().call(_this).then(() => {
        _this.state().isPersisted = true;
      });
    })();
  }

  loadPersistedValue(storageItemKey) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      var _getSharedStorageMana;

      if (!_this2.ready) return false;

      var _storageItemKey = storageItemKey !== null && storageItemKey !== void 0 ? storageItemKey : _this2._key;

      var loadedValue = yield (_getSharedStorageMana = getSharedStorageManager()) === null || _getSharedStorageMana === void 0 ? void 0 : _getSharedStorageMana.get(_storageItemKey, _this2.config.defaultStorageKey);
      if (loadedValue == null) return false;

      _this2.state().set(_this2.onMigrate != null ? _this2.onMigrate(loadedValue) : loadedValue, {
        storage: false,
        overwrite: true
      });

      _this2.setupSideEffects(_storageItemKey);

      return true;
    })();
  }

  persistValue(storageItemKey) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      if (!_this3.ready) return false;

      var _storageItemKey = storageItemKey !== null && storageItemKey !== void 0 ? storageItemKey : _this3._key;

      _this3.setupSideEffects(_storageItemKey);

      _this3.rebuildStorageSideEffect(_this3.state(), _storageItemKey);

      _this3.isPersisted = true;
      return true;
    })();
  }

  setupSideEffects(storageItemKey) {
    var _storageItemKey = storageItemKey !== null && storageItemKey !== void 0 ? storageItemKey : this._key;

    this.state().addSideEffect(StatePersistent.storeValueSideEffectKey, (instance, config) => {
      this.rebuildStorageSideEffect(this.state(), _storageItemKey, config);
    }, {
      weight: 0
    });
  }

  removePersistedValue(storageItemKey) {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      var _getSharedStorageMana2;

      if (!_this4.ready) return false;

      var _storageItemKey = storageItemKey || _this4._key;

      _this4.state().removeSideEffect(StatePersistent.storeValueSideEffectKey);

      (_getSharedStorageMana2 = getSharedStorageManager()) === null || _getSharedStorageMana2 === void 0 ? void 0 : _getSharedStorageMana2.remove(_storageItemKey, _this4.storageKeys);
      _this4.isPersisted = false;
      return true;
    })();
  }

  formatKey(key) {
    if (key == null && this.state()._key) return this.state()._key;
    if (key == null) return;
    if (this.state()._key == null) this.state()._key = key;
    return key;
  }

  rebuildStorageSideEffect(state, storageItemKey) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    if (config['storage'] == null || config.storage) {
      var _getSharedStorageMana3;

      (_getSharedStorageMana3 = getSharedStorageManager()) === null || _getSharedStorageMana3 === void 0 ? void 0 : _getSharedStorageMana3.set(storageItemKey, this.onSave != null ? this.onSave(this.state().getPersistableValue()) : this.state().getPersistableValue(), this.storageKeys);
    }
  }

}

_defineProperty(StatePersistent, "storeValueSideEffectKey", 'rebuildStateStorageValue');

class EnhancedState extends State {
  constructor(agileInstance, initialValue) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    super(agileInstance, initialValue, config);

    _defineProperty(this, "isPersisted", false);

    _defineProperty(this, "persistent", void 0);

    _defineProperty(this, "computeValueMethod", void 0);

    _defineProperty(this, "computeExistsMethod", void 0);

    _defineProperty(this, "currentInterval", void 0);

    this.computeExistsMethod = v => {
      return v != null;
    };
  }

  setKey(value) {
    var _this$persistent, _this$persistent2;

    var oldKey = this._key;
    super.setKey(value);
    if (value != null && ((_this$persistent = this.persistent) === null || _this$persistent === void 0 ? void 0 : _this$persistent._key) === oldKey) (_this$persistent2 = this.persistent) === null || _this$persistent2 === void 0 ? void 0 : _this$persistent2.setKey(value);
    return this;
  }

  undo() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    this.set(this.previousStateValue, config);
    return this;
  }

  reset() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    this.set(this.initialStateValue, config);
    return this;
  }

  patch(targetWithChanges) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      addNewProperties: true
    });

    if (!utils.isValidObject(this.nextStateValue, true)) {
      logCodeManager.log('14:03:02');
      return this;
    }

    if (!utils.isValidObject(targetWithChanges, true)) {
      logCodeManager.log('00:03:01', {
        replacers: ['TargetWithChanges', 'object']
      });
      return this;
    }

    if (Array.isArray(targetWithChanges) && Array.isArray(this.nextStateValue)) {
      this.nextStateValue = [...this.nextStateValue, ...targetWithChanges];
    } else {
      this.nextStateValue = utils.flatMerge(this.nextStateValue, targetWithChanges, {
        addNewProperties: config.addNewProperties
      });
    }

    this.ingest(config);
    return this;
  }

  watch(keyOrCallback, callback) {
    var generateKey = utils.isFunction(keyOrCallback);

    var _callback;

    var key;

    if (generateKey) {
      key = utils.generateId();
      _callback = keyOrCallback;
    } else {
      key = keyOrCallback;
      _callback = callback;
    }

    if (!utils.isFunction(_callback)) {
      logCodeManager.log('00:03:01', {
        replacers: ['Watcher Callback', 'function']
      });
      return this;
    }

    this.addSideEffect(key, instance => {
      _callback(instance.value, key);
    }, {
      weight: 0
    });
    return generateKey ? key : this;
  }

  removeWatcher(key) {
    this.removeSideEffect(key);
    return this;
  }

  onInaugurated(callback) {
    var watcherKey = 'InauguratedWatcherKey';
    this.watch(watcherKey, (value, key) => {
      callback(value, key);
      this.removeSideEffect(watcherKey);
    });
    return this;
  }

  interval(handler, delay) {
    if (!utils.isFunction(handler)) {
      logCodeManager.log('00:03:01', {
        replacers: ['Interval Callback', 'function']
      });
      return this;
    }

    if (this.currentInterval) {
      logCodeManager.log('14:03:03', {
        replacers: []
      }, this.currentInterval);
      return this;
    }

    this.currentInterval = setInterval(() => {
      this.set(handler(this._value));
    }, delay !== null && delay !== void 0 ? delay : 1000);
    return this;
  }

  clearInterval() {
    if (this.currentInterval) {
      clearInterval(this.currentInterval);
      delete this.currentInterval;
    }
  }

  get exists() {
    return !this.isPlaceholder && this.computeExistsMethod(this.value);
  }

  computeExists(method) {
    if (!utils.isFunction(method)) {
      logCodeManager.log('00:03:01', {
        replacers: ['Compute Exists Method', 'function']
      });
      return this;
    }

    this.computeExistsMethod = method;
    return this;
  }

  computeValue(method) {
    if (!utils.isFunction(method)) {
      logCodeManager.log('00:03:01', {
        replacers: ['Compute Value Method', 'function']
      });
      return this;
    }

    this.computeValueMethod = method;
    this.set(this.nextStateValue);
    return this;
  }

  is(value) {
    return utils.equal(value, this.value);
  }

  isNot(value) {
    return utils.notEqual(value, this.value);
  }

  invert() {
    switch (typeof this.nextStateValue) {
      case 'boolean':
        this.set(!this.nextStateValue);
        break;

      case 'object':
        if (Array.isArray(this.nextStateValue)) this.set(this.nextStateValue.reverse());
        break;

      case 'string':
        this.set(this.nextStateValue.split('').reverse().join(''));
        break;

      case 'number':
        this.set(this.nextStateValue * -1);
        break;

      default:
        logCodeManager.log('14:03:04', {
          replacers: [typeof this.nextStateValue]
        });
    }

    return this;
  }

  persist() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    config = utils.defineConfig(config, {
      key: this._key
    });
    if (this.persistent != null && this.isPersisted) return this;
    this.persistent = new StatePersistent(this, config);
    return this;
  }

  onLoad(callback) {
    if (!this.persistent) return this;

    if (!utils.isFunction(callback)) {
      logCodeManager.log('00:03:01', {
        replacers: ['OnLoad Callback', 'function']
      });
      return this;
    }

    this.persistent.onLoad = callback;
    if (this.isPersisted) callback(true);
    return this;
  }

  getPersistableValue() {
    return this._value;
  }

}

function createEnhancedState(initialValue) {
  var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  config = utils.defineConfig(config, {
    agileInstance: exports.sharedAgileInstance
  });
  return new EnhancedState(config.agileInstance, initialValue, config);
}

function createState(initialValue) {
  var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  return createEnhancedState(initialValue, config);
}

function createLightState(initialValue) {
  var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
  config = utils.defineConfig(config, {
    agileInstance: exports.sharedAgileInstance
  });
  return new State(config.agileInstance, initialValue, config);
}

class Computed extends State {
  constructor(agileInstance, computeFunction) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    super(agileInstance, Object.prototype.hasOwnProperty.call(config, 'initialValue') ? config.initialValue : !utils.isAsyncFunction(computeFunction) ? computeFunction() : null, {
      key: config.key,
      dependents: config.dependents
    });

    _defineProperty(this, "config", void 0);

    _defineProperty(this, "computeFunction", void 0);

    _defineProperty(this, "deps", new Set());

    _defineProperty(this, "hardCodedDeps", []);

    _defineProperty(this, "isComputed", true);

    config = utils.defineConfig(config, {
      computedDeps: [],
      autodetect: !utils.isAsyncFunction(computeFunction)
    });

    this.agileInstance = () => agileInstance;

    this.computeFunction = computeFunction;
    this.config = {
      autodetect: config.autodetect
    };
    this.hardCodedDeps = extractRelevantObservers(config.computedDeps).filter(dep => dep !== undefined);
    this.deps = new Set(this.hardCodedDeps);
    this.deps.forEach(observer => {
      observer.addDependent(this.observers['value']);
    });
    this.recompute({
      autodetect: config.autodetect,
      overwrite: true
    });
  }

  recompute() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    config = utils.defineConfig(config, {
      autodetect: false
    });
    this.compute({
      autodetect: config.autodetect
    }).then(result => {
      this.observers['value'].ingestValue(result, config);
    });
    return this;
  }

  updateComputeFunction(computeFunction) {
    var deps = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    config = utils.defineConfig(config, {
      autodetect: this.config.autodetect
    });
    this.deps.forEach(observer => {
      observer.removeDependent(this.observers['value']);
    });
    this.hardCodedDeps = extractRelevantObservers(deps).filter(dep => dep !== undefined);
    this.deps = new Set(this.hardCodedDeps);
    this.deps.forEach(observer => {
      observer.addDependent(this.observers['value']);
    });
    this.computeFunction = computeFunction;
    this.recompute(config);
    return this;
  }

  compute() {
    var _arguments = arguments,
        _this = this;

    return _asyncToGenerator(function* () {
      var config = _arguments.length > 0 && _arguments[0] !== undefined ? _arguments[0] : {};
      config = utils.defineConfig(config, {
        autodetect: _this.config.autodetect
      });
      if (config.autodetect) ComputedTracker.track();

      var computedValue = _this.computeFunction();

      if (config.autodetect) {
        var foundDeps = ComputedTracker.getTrackedObservers();

        _this.deps.forEach(observer => {
          if (!foundDeps.includes(observer) && !_this.hardCodedDeps.includes(observer)) {
            _this.deps.delete(observer);

            observer.removeDependent(_this.observers['value']);
          }
        });

        foundDeps.forEach(observer => {
          if (!_this.deps.has(observer)) {
            _this.deps.add(observer);

            observer.addDependent(_this.observers['value']);
          }
        });
      }

      return computedValue;
    })();
  }

}

function createComputed(computeFunction, configOrDeps) {
  var _config = {};

  if (Array.isArray(configOrDeps)) {
    _config = utils.defineConfig(_config, {
      computedDeps: configOrDeps
    });
  } else {
    if (configOrDeps) _config = configOrDeps;
  }

  _config = utils.defineConfig(_config, {
    agileInstance: exports.sharedAgileInstance
  });
  return new Computed(_config.agileInstance, computeFunction, _config);
}

class CollectionPersistent extends Persistent {
  constructor(collection) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    super(collection.agileInstance(), {
      loadValue: false
    });

    _defineProperty(this, "collection", void 0);

    config = utils.defineConfig(config, {
      loadValue: true,
      storageKeys: [],
      defaultStorageKey: null
    });

    this.collection = () => collection;

    this.instantiatePersistent({
      key: config.key,
      storageKeys: config.storageKeys,
      defaultStorageKey: config.defaultStorageKey
    });
    if (this.ready && config.loadValue) this.initialLoading();
  }

  initialLoading() {
    var _superprop_getInitialLoading = () => super.initialLoading,
        _this = this;

    return _asyncToGenerator(function* () {
      _superprop_getInitialLoading().call(_this).then(() => {
        _this.collection().isPersisted = true;
      });
    })();
  }

  loadPersistedValue(storageItemKey) {
    var _this2 = this;

    return _asyncToGenerator(function* () {
      var _getSharedStorageMana;

      if (!_this2.ready) return false;

      var _storageItemKey = storageItemKey !== null && storageItemKey !== void 0 ? storageItemKey : _this2._key;

      var isPersisted = yield (_getSharedStorageMana = getSharedStorageManager()) === null || _getSharedStorageMana === void 0 ? void 0 : _getSharedStorageMana.get(_storageItemKey, _this2.config.defaultStorageKey);
      if (!isPersisted) return false;

      var loadValuesIntoCollection = function () {
        var _ref = _asyncToGenerator(function* () {
          var _defaultGroup$persist;

          var defaultGroup = _this2.collection().getDefaultGroup();

          if (defaultGroup == null) return false;
          var defaultGroupStorageKey = CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
          defaultGroup.loadedInitialValue = false;
          defaultGroup.persist({
            key: defaultGroupStorageKey,
            loadValue: false,
            defaultStorageKey: _this2.config.defaultStorageKey || undefined,
            storageKeys: _this2.storageKeys,
            followCollectionPersistKeyPattern: false
          });
          if ((_defaultGroup$persist = defaultGroup.persistent) !== null && _defaultGroup$persist !== void 0 && _defaultGroup$persist.ready) yield defaultGroup.persistent.initialLoading();

          for (var itemKey of defaultGroup._value) {
            var item = _this2.collection().getItem(itemKey);

            var itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);

            if (item != null) {
              var _item$persistent;

              item.persist({
                key: itemStorageKey,
                loadValue: false,
                defaultStorageKey: _this2.config.defaultStorageKey || undefined,
                storageKeys: _this2.storageKeys,
                followCollectionPersistKeyPattern: false
              });
              if (item !== null && item !== void 0 && (_item$persistent = item.persistent) !== null && _item$persistent !== void 0 && _item$persistent.ready) yield item.persistent.initialLoading();
            } else {
              var _placeholderItem$pers;

              var placeholderItem = _this2.collection().getItemWithReference(itemKey);

              placeholderItem === null || placeholderItem === void 0 ? void 0 : placeholderItem.persist({
                key: itemStorageKey,
                loadValue: false,
                defaultStorageKey: _this2.config.defaultStorageKey,
                storageKeys: _this2.storageKeys,
                followCollectionPersistKeyPattern: false
              });

              if (placeholderItem !== null && placeholderItem !== void 0 && (_placeholderItem$pers = placeholderItem.persistent) !== null && _placeholderItem$pers !== void 0 && _placeholderItem$pers.ready) {
                var loadedPersistedValueIntoItem = yield placeholderItem.persistent.loadPersistedValue();

                if (loadedPersistedValueIntoItem) {
                  _this2.collection().assignItem(placeholderItem, {
                    overwrite: true,
                    rebuildGroups: false
                  });

                  placeholderItem.isPersisted = true;
                  _this2.collection().size += 1;
                }
              }
            }
          }

          defaultGroup.loadedInitialValue = true;
          return true;
        });

        return function loadValuesIntoCollection() {
          return _ref.apply(this, arguments);
        };
      }();

      var success = yield loadValuesIntoCollection();
      if (success) _this2.setupSideEffects(_storageItemKey);
      return success;
    })();
  }

  persistValue(storageItemKey) {
    var _this3 = this;

    return _asyncToGenerator(function* () {
      var _getSharedStorageMana2;

      if (!_this3.ready) return false;

      var _storageItemKey = storageItemKey !== null && storageItemKey !== void 0 ? storageItemKey : _this3._key;

      var defaultGroup = _this3.collection().getDefaultGroup();

      if (defaultGroup == null) return false;
      var defaultGroupStorageKey = CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
      (_getSharedStorageMana2 = getSharedStorageManager()) === null || _getSharedStorageMana2 === void 0 ? void 0 : _getSharedStorageMana2.set(_storageItemKey, true, _this3.storageKeys);
      defaultGroup.persist({
        key: defaultGroupStorageKey,
        defaultStorageKey: _this3.config.defaultStorageKey || undefined,
        storageKeys: _this3.storageKeys,
        followCollectionPersistKeyPattern: false
      });

      for (var itemKey of defaultGroup._value) {
        var item = _this3.collection().getItem(itemKey);

        var itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
        item === null || item === void 0 ? void 0 : item.persist({
          key: itemStorageKey,
          defaultStorageKey: _this3.config.defaultStorageKey || undefined,
          storageKeys: _this3.storageKeys,
          followCollectionPersistKeyPattern: false
        });
      }

      _this3.setupSideEffects(_storageItemKey);

      _this3.isPersisted = true;
      return true;
    })();
  }

  setupSideEffects(storageItemKey) {
    var _storageItemKey = storageItemKey !== null && storageItemKey !== void 0 ? storageItemKey : this._key;

    var defaultGroup = this.collection().getDefaultGroup();
    if (defaultGroup == null) return;
    defaultGroup.addSideEffect(CollectionPersistent.defaultGroupSideEffectKey, instance => this.rebuildStorageSideEffect(instance, _storageItemKey), {
      weight: 0
    });
  }

  removePersistedValue(storageItemKey) {
    var _this4 = this;

    return _asyncToGenerator(function* () {
      var _getSharedStorageMana3, _defaultGroup$persist2;

      if (!_this4.ready) return false;

      var _storageItemKey = storageItemKey !== null && storageItemKey !== void 0 ? storageItemKey : _this4._key;

      var defaultGroup = _this4.collection().getDefaultGroup();

      if (!defaultGroup) return false;
      var defaultGroupStorageKey = CollectionPersistent.getGroupStorageKey(defaultGroup._key, _storageItemKey);
      (_getSharedStorageMana3 = getSharedStorageManager()) === null || _getSharedStorageMana3 === void 0 ? void 0 : _getSharedStorageMana3.remove(_storageItemKey, _this4.storageKeys);
      (_defaultGroup$persist2 = defaultGroup.persistent) === null || _defaultGroup$persist2 === void 0 ? void 0 : _defaultGroup$persist2.removePersistedValue(defaultGroupStorageKey);
      defaultGroup.removeSideEffect(CollectionPersistent.defaultGroupSideEffectKey);

      for (var itemKey of defaultGroup._value) {
        var _item$persistent2;

        var item = _this4.collection().getItem(itemKey);

        var itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
        item === null || item === void 0 ? void 0 : (_item$persistent2 = item.persistent) === null || _item$persistent2 === void 0 ? void 0 : _item$persistent2.removePersistedValue(itemStorageKey);
      }

      _this4.isPersisted = false;
      return true;
    })();
  }

  formatKey(key) {
    if (key == null && this.collection()._key) return this.collection()._key;
    if (key == null) return;
    if (this.collection()._key == null) this.collection()._key = key;
    return key;
  }

  rebuildStorageSideEffect(group, storageItemKey) {
    var _collection$persisten;

    var collection = group.collection();

    var _storageItemKey = storageItemKey || ((_collection$persisten = collection.persistent) === null || _collection$persisten === void 0 ? void 0 : _collection$persisten._key);

    if (group.previousStateValue.length === group._value.length) return;

    var addedKeys = group._value.filter(key => !group.previousStateValue.includes(key));

    var removedKeys = group.previousStateValue.filter(key => !group._value.includes(key));
    addedKeys.forEach(itemKey => {
      var item = collection.getItem(itemKey);
      var itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
      if (item != null && !item.isPersisted) item.persist({
        key: itemStorageKey,
        defaultStorageKey: this.config.defaultStorageKey || undefined,
        storageKeys: this.storageKeys,
        followCollectionPersistKeyPattern: false
      });
    });
    removedKeys.forEach(itemKey => {
      var _item$persistent3;

      var item = collection.getItem(itemKey);
      var itemStorageKey = CollectionPersistent.getItemStorageKey(itemKey, _storageItemKey);
      if (item != null && item.isPersisted) (_item$persistent3 = item.persistent) === null || _item$persistent3 === void 0 ? void 0 : _item$persistent3.removePersistedValue(itemStorageKey);
    });
  }

  static getItemStorageKey(itemKey, collectionKey) {
    if (itemKey == null || collectionKey == null) logCodeManager.log('1A:02:00');
    if (itemKey == null) itemKey = 'unknown';
    if (collectionKey == null) collectionKey = 'unknown';
    return this.storageItemKeyPattern.replace('${collectionKey}', collectionKey.toString()).replace('${itemKey}', itemKey.toString());
  }

  static getGroupStorageKey(groupKey, collectionKey) {
    if (groupKey == null || collectionKey == null) logCodeManager.log('1A:02:01');
    if (groupKey == null) groupKey = 'unknown';
    if (collectionKey == null) collectionKey = 'unknown';
    return this.storageGroupKeyPattern.replace('${collectionKey}', collectionKey.toString()).replace('${groupKey}', groupKey.toString());
  }

}

_defineProperty(CollectionPersistent, "defaultGroupSideEffectKey", 'rebuildGroupStorageValue');

_defineProperty(CollectionPersistent, "storageItemKeyPattern", '_${collectionKey}_item_${itemKey}');

_defineProperty(CollectionPersistent, "storageGroupKeyPattern", '_${collectionKey}_group_${groupKey}');

class Item extends EnhancedState {
  constructor(collection, data) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    super(collection.agileInstance(), data, {
      isPlaceholder: config.isPlaceholder,
      key: data[collection.config.primaryKey]
    });

    _defineProperty(this, "collection", void 0);

    _defineProperty(this, "selectedBy", new Set());

    this.collection = () => collection;

    if (this._key != null) {
      this.addRebuildGroupThatIncludeItemKeySideEffect(this._key);
    }
  }

  setKey(value) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    super.setKey(value);
    config = utils.defineConfig(config, {
      sideEffects: {
        enabled: true,
        exclude: []
      },
      background: false,
      force: false,
      storage: true,
      overwrite: false
    });
    if (value == null) return this;
    this.removeSideEffect(Item.updateGroupSideEffectKey);
    this.addRebuildGroupThatIncludeItemKeySideEffect(value);
    this.patch({
      [this.collection().config.primaryKey]: value
    }, config);
    return this;
  }

  persist() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    config = utils.defineConfig(config, {
      key: this._key,
      followCollectionPersistKeyPattern: true
    });

    if (config.followCollectionPersistKeyPattern) {
      config.key = CollectionPersistent.getItemStorageKey(config.key || this._key, this.collection()._key);
    }

    super.persist(config);
    return this;
  }

  addRebuildGroupThatIncludeItemKeySideEffect(itemKey) {
    this.addSideEffect(Item.updateGroupSideEffectKey, (instance, config) => {
      instance.collection().rebuildGroupsThatIncludeItemKey(itemKey, config);
    }, {
      weight: 100
    });
  }

}

_defineProperty(Item, "updateGroupSideEffectKey", 'rebuildGroup');

class Selector extends EnhancedState {
  constructor(collection, itemKey) {
    var _config;

    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    config = utils.defineConfig(config, {
      isPlaceholder: false
    });
    super(collection.agileInstance(), null, config);

    _defineProperty(this, "collection", void 0);

    _defineProperty(this, "_item", void 0);

    _defineProperty(this, "_itemKey", void 0);

    this.collection = () => collection;

    this._item = null;
    this._itemKey = !config.isPlaceholder && itemKey != null ? itemKey : null;
    this._key = (_config = config) === null || _config === void 0 ? void 0 : _config.key;
    this.isPlaceholder = true;
    if (this._itemKey != null) this.select(itemKey, {
      overwrite: true
    });
  }

  get itemKey() {
    return this._itemKey;
  }

  set itemKey(value) {
    this.select(value);
  }

  get item() {
    return this._item;
  }

  set item(value) {
    if (value !== null && value !== void 0 && value._key) this.select(value._key);
  }

  select(itemKey) {
    var _this$_item$isPlaceho, _this$_item;

    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      background: false,
      sideEffects: {
        enabled: true,
        exclude: []
      },
      force: false,
      overwrite: (_this$_item$isPlaceho = (_this$_item = this._item) === null || _this$_item === void 0 ? void 0 : _this$_item.isPlaceholder) !== null && _this$_item$isPlaceho !== void 0 ? _this$_item$isPlaceho : false,
      storage: true
    });
    if ((!this.collection().isInstantiated || this.hasSelected(itemKey)) && !config.force) return this;
    this.unselect({
      background: itemKey != null
    });
    if (itemKey == null) return this;
    var newItem = this.collection().getItemWithReference(itemKey);
    this._itemKey = itemKey;
    this._item = newItem;
    newItem.selectedBy.add(this._key);
    newItem.addSideEffect(Selector.rebuildSelectorSideEffectKey, (instance, config) => this.rebuildSelector(config), {
      weight: 100
    });
    this.addSideEffect(Selector.rebuildItemSideEffectKey, (instance, config) => {
      var _instance$_item, _instance$_item2;

      if (!((_instance$_item = instance._item) !== null && _instance$_item !== void 0 && _instance$_item.isPlaceholder)) (_instance$_item2 = instance._item) === null || _instance$_item2 === void 0 ? void 0 : _instance$_item2.set(instance._value, _objectSpread2(_objectSpread2({}, config), {
        sideEffects: {
          enabled: true,
          exclude: [Selector.rebuildSelectorSideEffectKey]
        }
      }));
    }, {
      weight: 90
    });
    this.rebuildSelector(config);
    return this;
  }

  reselect() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    if (this._itemKey != null && !this.hasSelected(this._itemKey) || config.force) this.select(this._itemKey, config);
    return this;
  }

  unselect() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var item = this.collection().getItem(this._itemKey, {
      notExisting: true
    });

    if (item != null) {
      item.selectedBy.delete(this._key);
      item.removeSideEffect(Selector.rebuildSelectorSideEffectKey);
      item.removeSideEffect(Selector.rebuildItemSideEffectKey);
      if (item.isPlaceholder && this._itemKey != null) delete this.collection().data[this._itemKey];
    }

    this._item = null;
    this._itemKey = null;
    this.rebuildSelector(config);
    this.isPlaceholder = true;
    return this;
  }

  hasSelected(itemKey) {
    var correctlySelected = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : true;

    if (correctlySelected) {
      return this._itemKey === itemKey && this._item != null && this._item.selectedBy.has(this._key);
    }

    return this._itemKey === itemKey;
  }

  rebuildSelector() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};

    if (this._item == null || this._item.isPlaceholder) {
      this.set(null, config);
      return this;
    }

    this.set(this._item._value, config);
    return this;
  }

}

_defineProperty(Selector, "rebuildSelectorSideEffectKey", 'rebuildSelector');

_defineProperty(Selector, "rebuildItemSideEffectKey", 'rebuildItem');

class GroupObserver extends Observer {
  constructor(group) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    super(group.agileInstance(), utils.defineConfig(config, {
      value: utils.copy(group._output)
    }));

    _defineProperty(this, "group", void 0);

    _defineProperty(this, "nextGroupValue", void 0);

    this.group = () => group;

    this.nextGroupValue = utils.copy(group._output);
  }

  ingest() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    this.ingestOutput(this.group().nextGroupOutput, config);
  }

  ingestOutput(newGroupOutput) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    var group = this.group();
    config = utils.defineConfig(config, {
      perform: true,
      force: false,
      key: logCodeManager.allowLogging ? "".concat(this.key != null ? this.key + '_' : '').concat(utils.generateId(), "_output") : undefined
    });

    if (group.isPlaceholder) {
      config.force = true;
    }

    this.nextGroupValue = utils.copy(newGroupOutput);
    if (utils.equal(group._output, this.nextGroupValue) && !config.force) return;
    var job = new RuntimeJob(this, config);
    this.agileInstance().runtime.ingest(job, {
      perform: config.perform
    });
  }

  perform(job) {
    var observer = job.observer;
    var group = observer.group();
    group._output = utils.copy(observer.nextGroupValue);
    group.nextGroupOutput = utils.copy(observer.nextGroupValue);
    observer.previousValue = Object.freeze(utils.copy(observer.value));
    observer.value = utils.copy(group._output);
  }

}

class Group extends EnhancedState {
  constructor(collection) {
    var initialItems = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    super(collection.agileInstance(), initialItems, config);

    _defineProperty(this, "collection", void 0);

    _defineProperty(this, "_output", []);

    _defineProperty(this, "nextGroupOutput", []);

    _defineProperty(this, "_preciseItemKeys", []);

    _defineProperty(this, "observers", {});

    _defineProperty(this, "notFoundItemKeys", []);

    _defineProperty(this, "loadedInitialValue", true);

    this.observers['value'] = new StateObserver(this, {
      key: config.key
    });
    this.observers['output'] = new GroupObserver(this, {
      key: config.key
    });

    this.collection = () => collection;

    this.addSideEffect(Group.rebuildGroupSideEffectKey, (state, config) => {
      var _config$any;

      this.rebuild((config === null || config === void 0 ? void 0 : (_config$any = config.any) === null || _config$any === void 0 ? void 0 : _config$any.trackedChanges) || [], config);
    });
    this.rebuild();
  }

  get output() {
    ComputedTracker.tracked(this.observers['output']);
    return utils.copy(this._output);
  }

  set output(value) {
    logCodeManager.log('1C:03:00', {
      replacers: [this._key]
    });
  }

  has(itemKey) {
    return this.value.indexOf(itemKey) !== -1;
  }

  get size() {
    return this.value.length;
  }

  remove(itemKeys) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    var _itemKeys = utils.normalizeArray(itemKeys);

    var notExistingItemKeysInCollection = [];
    var notExistingItemKeys = [];
    var newGroupValue = utils.copy(this.nextStateValue);
    var updatedPreciseItemKeys = utils.copy(this._preciseItemKeys);
    config = utils.defineConfig(config, {
      softRebuild: true,
      any: {}
    });
    config.any['trackedChanges'] = [];

    _itemKeys.forEach(itemKey => {
      var exists = newGroupValue.includes(itemKey);

      if (!exists) {
        notExistingItemKeys.push(itemKey);
        notExistingItemKeysInCollection.push(itemKey);
        return;
      }

      if (config.softRebuild) {
        var index = updatedPreciseItemKeys.findIndex(ik => ik === itemKey);

        if (index !== -1) {
          updatedPreciseItemKeys.splice(index, 1);
          config.any['trackedChanges'].push({
            index,
            method: exports.TrackedChangeMethod.REMOVE,
            key: itemKey
          });
        }
      }

      if (!this.collection().getItem(itemKey)) notExistingItemKeysInCollection.push(itemKey);
      newGroupValue = newGroupValue.filter(key => key !== itemKey);
    });

    if (notExistingItemKeys.length >= _itemKeys.length) return this;
    if (notExistingItemKeysInCollection.length >= _itemKeys.length) config.background = true;
    this.set(newGroupValue, config);
    return this;
  }

  add(itemKeys) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    var _itemKeys = utils.normalizeArray(itemKeys);

    var notExistingItemKeysInCollection = [];
    var existingItemKeys = [];
    var newGroupValue = utils.copy(this.nextStateValue);
    var updatedPreciseItemKeys = utils.copy(this._preciseItemKeys);
    config = utils.defineConfig(config, {
      method: 'push',
      softRebuild: true,
      any: {}
    });
    config.any['trackedChanges'] = [];

    _itemKeys.forEach(itemKey => {
      var exists = newGroupValue.includes(itemKey);
      if (!this.collection().getItem(itemKey)) notExistingItemKeysInCollection.push(itemKey);

      if (exists) {
        existingItemKeys.push(itemKey);
        return;
      }

      if (config.softRebuild) {
        var index = config.method === 'push' ? updatedPreciseItemKeys.length : 0;
        updatedPreciseItemKeys.push(itemKey);
        config.any['trackedChanges'].push({
          index,
          method: exports.TrackedChangeMethod.ADD,
          key: itemKey
        });
      }

      newGroupValue[config.method || 'push'](itemKey);
    });

    if (existingItemKeys.length >= _itemKeys.length) return this;
    if (notExistingItemKeysInCollection.concat(existingItemKeys).length >= _itemKeys.length) config.background = true;
    this.set(newGroupValue, config);
    return this;
  }

  replace(oldItemKey, newItemKey) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    var newGroupValue = utils.copy(this._value);
    newGroupValue.splice(newGroupValue.indexOf(oldItemKey), 1, newItemKey);
    this.set(newGroupValue, config);
    return this;
  }

  getItems() {
    return this.value.map(itemKey => this.collection().getItem(itemKey)).filter(item => item !== undefined);
  }

  persist() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    config = utils.defineConfig(config, {
      key: this._key,
      followCollectionPersistKeyPattern: true
    });

    if (config.followCollectionPersistKeyPattern) {
      config.key = CollectionPersistent.getGroupStorageKey(config.key || this._key, this.collection()._key);
    }

    super.persist(config);
    return this;
  }

  rebuild() {
    var trackedChanges = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : [];
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    if (!this.collection().isInstantiated) return this;
    var notFoundItemKeys = [];

    if (trackedChanges.length > 0) {
      trackedChanges.forEach(change => {
        var item = this.collection().getItem(change.key);

        switch (change.method) {
          case exports.TrackedChangeMethod.ADD:
            if (item != null) {
              this._preciseItemKeys.splice(change.index, 0, change.key);

              this.nextGroupOutput.splice(change.index, 0, utils.copy(item._value));
            } else {
              notFoundItemKeys.push(change.key);
            }

            break;

          case exports.TrackedChangeMethod.UPDATE:
            if (item != null) {
              this.nextGroupOutput[change.index] = utils.copy(item._value);
            } else {
              notFoundItemKeys.push(change.key);
            }

            break;

          case exports.TrackedChangeMethod.REMOVE:
            this._preciseItemKeys.splice(change.index, 1);

            this.nextGroupOutput.splice(change.index, 1);
            break;
        }
      });
      this.observers['output'].ingest(config);
    } else {
      var groupItemValues = [];
      this._preciseItemKeys = [];

      this._value.forEach(itemKey => {
        var item = this.collection().getItem(itemKey);

        if (item != null) {
          groupItemValues.push(item._value);

          this._preciseItemKeys.push(itemKey);
        } else notFoundItemKeys.push(itemKey);
      });

      this.observers['output'].ingestOutput(groupItemValues, config);
    }

    if (notFoundItemKeys.length > 0 && this.loadedInitialValue) {
      logCodeManager.log('1C:02:00', {
        replacers: [this.collection()._key, this._key]
      }, notFoundItemKeys);
    }

    this.notFoundItemKeys = notFoundItemKeys;
    return this;
  }

}

_defineProperty(Group, "rebuildGroupSideEffectKey", 'rebuildGroup');

exports.TrackedChangeMethod = void 0;

(function (TrackedChangeMethod) {
  TrackedChangeMethod[TrackedChangeMethod["ADD"] = 0] = "ADD";
  TrackedChangeMethod[TrackedChangeMethod["REMOVE"] = 1] = "REMOVE";
  TrackedChangeMethod[TrackedChangeMethod["UPDATE"] = 2] = "UPDATE";
})(exports.TrackedChangeMethod || (exports.TrackedChangeMethod = {}));

class Collection {
  constructor(agileInstance) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    _defineProperty(this, "agileInstance", void 0);

    _defineProperty(this, "config", void 0);

    _defineProperty(this, "initialConfig", void 0);

    _defineProperty(this, "_key", void 0);

    _defineProperty(this, "size", 0);

    _defineProperty(this, "data", {});

    _defineProperty(this, "isPersisted", false);

    _defineProperty(this, "persistent", void 0);

    _defineProperty(this, "groups", {});

    _defineProperty(this, "selectors", {});

    _defineProperty(this, "isInstantiated", false);

    _defineProperty(this, "isCollection", true);

    this.agileInstance = () => agileInstance;

    var _config = typeof config === 'function' ? config(this) : config;

    _config = utils.defineConfig(_config, {
      primaryKey: 'id',
      groups: {},
      selectors: {},
      defaultGroupKey: 'default'
    });
    this._key = _config.key;
    this.config = {
      defaultGroupKey: _config.defaultGroupKey,
      primaryKey: _config.primaryKey
    };
    this.initialConfig = _config;
    this.initGroups(_config.groups);
    this.initSelectors(_config.selectors);
    this.isInstantiated = true;
    if (_config.initialData) this.collect(_config.initialData);

    for (var key in this.selectors) {
      this.selectors[key].reselect();
    }
  }

  set key(value) {
    this.setKey(value);
  }

  get key() {
    return this._key;
  }

  setKey(value) {
    var _this$persistent, _this$persistent2;

    var oldKey = this._key;
    this._key = value;
    if (value != null && ((_this$persistent = this.persistent) === null || _this$persistent === void 0 ? void 0 : _this$persistent._key) === oldKey) (_this$persistent2 = this.persistent) === null || _this$persistent2 === void 0 ? void 0 : _this$persistent2.setKey(value);
    return this;
  }

  Group(initialItems) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    if (this.isInstantiated) {
      var _config$key;

      var key = (_config$key = config.key) !== null && _config$key !== void 0 ? _config$key : utils.generateId();
      logCodeManager.log('1B:02:00');
      return this.createGroup(key, initialItems);
    }

    return new Group(this, initialItems, config);
  }

  Selector(initialKey) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    if (this.isInstantiated) {
      var _config$key2;

      var key = (_config$key2 = config.key) !== null && _config$key2 !== void 0 ? _config$key2 : utils.generateId();
      logCodeManager.log('1B:02:01');
      return this.createSelector(key, initialKey);
    }

    return new Selector(this, initialKey, config);
  }

  initGroups(groups) {
    if (!groups) return;
    var groupsObject = {};

    if (Array.isArray(groups)) {
      groups.forEach(groupKey => {
        groupsObject[groupKey] = new Group(this, [], {
          key: groupKey
        });
      });
    } else groupsObject = groups;

    groupsObject[this.config.defaultGroupKey] = new Group(this, [], {
      key: this.config.defaultGroupKey
    });

    for (var key in groupsObject) {
      if (groupsObject[key]._key == null) groupsObject[key].setKey(key);
    }

    this.groups = groupsObject;
  }

  initSelectors(selectors) {
    if (!selectors) return;
    var selectorsObject = {};

    if (Array.isArray(selectors)) {
      selectors.forEach(selectorKey => {
        selectorsObject[selectorKey] = new Selector(this, selectorKey, {
          key: selectorKey
        });
      });
    } else selectorsObject = selectors;

    for (var key in selectorsObject) {
      if (selectorsObject[key]._key == null) selectorsObject[key].setKey(key);
    }

    this.selectors = selectorsObject;
  }

  collect(data, groupKeys) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    var _data = utils.normalizeArray(data);

    var _groupKeys = utils.normalizeArray(groupKeys);

    var defaultGroupKey = this.config.defaultGroupKey;
    var primaryKey = this.config.primaryKey;
    config = utils.defineConfig(config, {
      method: 'push',
      background: false,
      patch: false,
      select: false
    });
    if (!_groupKeys.includes(defaultGroupKey)) _groupKeys.push(defaultGroupKey);

    _groupKeys.forEach(key => this.groups[key] == null && this.createGroup(key));

    _data.forEach((data, index) => {
      var itemKey;
      var success = false;

      if (data instanceof Item) {
        success = this.assignItem(data, {
          background: config.background
        });
        itemKey = data._key;
      } else {
        success = this.assignData(data, {
          patch: config.patch,
          background: config.background
        });
        itemKey = data[primaryKey];
      }

      if (success) {
        _groupKeys.forEach(groupKey => {
          var _this$getGroup;

          (_this$getGroup = this.getGroup(groupKey)) === null || _this$getGroup === void 0 ? void 0 : _this$getGroup.add(itemKey, {
            method: config.method,
            background: config.background
          });
        });

        if (config.select) this.createSelector(itemKey, itemKey);
      }

      if (config.forEachItem) config.forEachItem(data, itemKey, success, index);
    });

    return this;
  }

  update(itemKey, changes) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    var item = this.getItem(itemKey, {
      notExisting: true
    });
    var primaryKey = this.config.primaryKey;
    config = utils.defineConfig(config, {
      patch: true,
      background: false
    });

    if (item == null) {
      logCodeManager.log('1B:03:00', {
        replacers: [itemKey, this._key]
      });
      return undefined;
    }

    if (!utils.isValidObject(changes)) {
      logCodeManager.log('1B:03:01', {
        replacers: [itemKey, this._key]
      });
      return undefined;
    }

    var oldItemKey = item._value[primaryKey];
    var newItemKey = changes[primaryKey] || oldItemKey;
    if (oldItemKey !== newItemKey) this.updateItemKey(oldItemKey, newItemKey, {
      background: config.background
    });

    if (config.patch) {
      if (changes[primaryKey]) delete changes[primaryKey];
      var patchConfig = typeof config.patch === 'object' ? config.patch : {};
      patchConfig = _objectSpread2({
        addNewProperties: true
      }, patchConfig);
      item.patch(changes, {
        background: config.background,
        addNewProperties: patchConfig.addNewProperties
      });
    } else {
      if (changes[this.config.primaryKey] !== itemKey) {
        changes[this.config.primaryKey] = itemKey;
        logCodeManager.log('1B:02:02', {}, changes);
      }

      item.set(changes, {
        background: config.background
      });
    }

    return item;
  }

  createGroup(groupKey) {
    var initialItems = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : [];
    var group = this.getGroup(groupKey, {
      notExisting: true
    });
    if (!this.isInstantiated) logCodeManager.log('1B:02:03');

    if (group != null) {
      if (!group.isPlaceholder) {
        logCodeManager.log('1B:03:02', {
          replacers: [groupKey]
        });
        return group;
      }

      group.set(initialItems, {
        overwrite: true
      });
      return group;
    }

    group = new Group(this, initialItems, {
      key: groupKey
    });
    this.groups[groupKey] = group;
    return group;
  }

  hasGroup(groupKey) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    return !!this.getGroup(groupKey, config);
  }

  getGroup(groupKey) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      notExisting: false
    });
    var group = groupKey ? this.groups[groupKey] : undefined;
    if (group == null || !config.notExisting && !group.exists) return undefined;
    ComputedTracker.tracked(group.observers['value']);
    return group;
  }

  getDefaultGroup() {
    return this.getGroup(this.config.defaultGroupKey);
  }

  getGroupWithReference(groupKey) {
    var group = this.getGroup(groupKey, {
      notExisting: true
    });

    if (group == null) {
      group = new Group(this, [], {
        key: groupKey,
        isPlaceholder: true
      });
      this.groups[groupKey] = group;
    }

    ComputedTracker.tracked(group.observers['value']);
    return group;
  }

  removeGroup(groupKey) {
    if (this.groups[groupKey] != null) delete this.groups[groupKey];
    return this;
  }

  getGroupCount() {
    var size = 0;
    Object.keys(this.groups).map(() => size++);
    return size;
  }

  createSelector(selectorKey, itemKey) {
    var selector = this.getSelector(selectorKey, {
      notExisting: true
    });
    if (!this.isInstantiated) logCodeManager.log('1B:02:04');

    if (selector != null) {
      if (!selector.isPlaceholder) {
        logCodeManager.log('1B:03:03', {
          replacers: [selectorKey]
        });
        return selector;
      }

      selector.select(itemKey, {
        overwrite: true
      });
      return selector;
    }

    selector = new Selector(this, itemKey, {
      key: selectorKey
    });
    this.selectors[selectorKey] = selector;
    return selector;
  }

  select(itemKey) {
    return this.createSelector(itemKey, itemKey);
  }

  hasSelector(selectorKey) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    return !!this.getSelector(selectorKey, config);
  }

  getSelector(selectorKey) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      notExisting: false
    });
    var selector = selectorKey ? this.selectors[selectorKey] : undefined;
    if (selector == null || !config.notExisting && !selector.exists) return undefined;
    ComputedTracker.tracked(selector.observers['value']);
    return selector;
  }

  getSelectorWithReference(selectorKey) {
    var selector = this.getSelector(selectorKey, {
      notExisting: true
    });

    if (selector == null) {
      selector = new Selector(this, null, {
        key: selectorKey,
        isPlaceholder: true
      });
      this.selectors[selectorKey] = selector;
    }

    ComputedTracker.tracked(selector.observers['value']);
    return selector;
  }

  removeSelector(selectorKey) {
    if (this.selectors[selectorKey] != null) {
      this.selectors[selectorKey].unselect();
      delete this.selectors[selectorKey];
    }

    return this;
  }

  getSelectorCount() {
    var size = 0;
    Object.keys(this.selectors).map(() => size++);
    return size;
  }

  hasItem(itemKey) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    return !!this.getItem(itemKey, config);
  }

  getItem(itemKey) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      notExisting: false
    });
    var item = itemKey != null ? this.data[itemKey] : undefined;
    if (item == null || !config.notExisting && !item.exists) return undefined;
    ComputedTracker.tracked(item.observers['value']);
    return item;
  }

  getItemWithReference(itemKey) {
    var item = this.getItem(itemKey, {
      notExisting: true
    });
    if (item == null) item = this.createPlaceholderItem(itemKey, true);
    ComputedTracker.tracked(item.observers['value']);
    return item;
  }

  createPlaceholderItem(itemKey) {
    var addToCollection = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : false;
    var item = new Item(this, {
      [this.config.primaryKey]: itemKey,
      dummy: 'item'
    }, {
      isPlaceholder: true
    });
    if (addToCollection && !Object.prototype.hasOwnProperty.call(this.data, itemKey)) this.data[itemKey] = item;
    ComputedTracker.tracked(item.observers['value']);
    return item;
  }

  getItemValue(itemKey) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    var item = this.getItem(itemKey, config);
    if (item == null) return undefined;
    return item.value;
  }

  getAllItems() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    config = utils.defineConfig(config, {
      notExisting: false
    });
    var defaultGroup = this.getDefaultGroup();
    var items = [];

    if (config.notExisting) {
      for (var key in this.data) {
        items.push(this.data[key]);
      }
    } else {
      items = (defaultGroup === null || defaultGroup === void 0 ? void 0 : defaultGroup.getItems()) || [];
    }

    return items;
  }

  getAllItemValues() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    var items = this.getAllItems(config);
    return items.map(item => item.value);
  }

  persist() {
    var config = arguments.length > 0 && arguments[0] !== undefined ? arguments[0] : {};
    config = utils.defineConfig(config, {
      key: this.key
    });
    if (this.persistent != null && this.isPersisted) return this;
    this.persistent = new CollectionPersistent(this, config);
    return this;
  }

  onLoad(callback) {
    if (!this.persistent) return this;

    if (!utils.isFunction(callback)) {
      logCodeManager.log('00:03:01', {
        replacers: ['OnLoad Callback', 'function']
      });
      return this;
    }

    this.persistent.onLoad = callback;
    if (this.isPersisted) callback(true);
    return this;
  }

  reset() {
    this.data = {};
    this.size = 0;

    for (var key in this.groups) {
      var _this$getGroup2;

      (_this$getGroup2 = this.getGroup(key)) === null || _this$getGroup2 === void 0 ? void 0 : _this$getGroup2.reset();
    }

    for (var _key in this.selectors) {
      var _this$getSelector;

      (_this$getSelector = this.getSelector(_key)) === null || _this$getSelector === void 0 ? void 0 : _this$getSelector.reset();
    }

    return this;
  }

  put(itemKeys, groupKeys) {
    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};

    var _itemKeys = utils.normalizeArray(itemKeys);

    var _groupKeys = utils.normalizeArray(groupKeys);

    _groupKeys.forEach(groupKey => {
      var _this$getGroup3;

      (_this$getGroup3 = this.getGroup(groupKey)) === null || _this$getGroup3 === void 0 ? void 0 : _this$getGroup3.add(_itemKeys, config);
    });

    return this;
  }

  move(itemKeys, oldGroupKey, newGroupKey) {
    var _this$getGroup4, _this$getGroup5;

    var config = arguments.length > 3 && arguments[3] !== undefined ? arguments[3] : {};

    var _itemKeys = utils.normalizeArray(itemKeys);

    (_this$getGroup4 = this.getGroup(oldGroupKey)) === null || _this$getGroup4 === void 0 ? void 0 : _this$getGroup4.remove(_itemKeys, config);
    (_this$getGroup5 = this.getGroup(newGroupKey)) === null || _this$getGroup5 === void 0 ? void 0 : _this$getGroup5.add(_itemKeys, config);
    return this;
  }

  updateItemKey(oldItemKey, newItemKey) {
    var _item$persistent;

    var config = arguments.length > 2 && arguments[2] !== undefined ? arguments[2] : {};
    var item = this.getItem(oldItemKey, {
      notExisting: true
    });
    config = utils.defineConfig(config, {
      background: false
    });
    if (item == null || oldItemKey === newItemKey) return false;

    if (this.hasItem(newItemKey)) {
      logCodeManager.log('1B:03:04', {
        replacers: [oldItemKey, newItemKey, this._key]
      });
      return false;
    }

    delete this.data[oldItemKey];
    this.data[newItemKey] = item;
    item.setKey(newItemKey, {
      background: config.background
    });
    if (item.persistent != null && item.persistent._key === CollectionPersistent.getItemStorageKey(oldItemKey, this._key)) (_item$persistent = item.persistent) === null || _item$persistent === void 0 ? void 0 : _item$persistent.setKey(CollectionPersistent.getItemStorageKey(newItemKey, this._key));

    for (var groupKey in this.groups) {
      var group = this.getGroup(groupKey, {
        notExisting: true
      });
      if (group == null || !group.has(oldItemKey)) continue;
      group.replace(oldItemKey, newItemKey, {
        background: config.background
      });
    }

    for (var selectorKey in this.selectors) {
      var selector = this.getSelector(selectorKey, {
        notExisting: true
      });
      if (selector == null) continue;

      if (selector.hasSelected(newItemKey, false)) {
        selector.reselect({
          force: true,
          background: config.background
        });
      }

      if (selector.hasSelected(oldItemKey, false)) selector.select(newItemKey, {
        background: config.background
      });
    }

    return true;
  }

  getGroupKeysThatHaveItemKey(itemKey) {
    var groupKeys = [];

    for (var groupKey in this.groups) {
      var group = this.groups[groupKey];
      if (group !== null && group !== void 0 && group.has(itemKey)) groupKeys.push(groupKey);
    }

    return groupKeys;
  }

  remove(itemKeys) {
    return {
      fromGroups: groups => this.removeFromGroups(itemKeys, groups),
      everywhere: config => this.removeItems(itemKeys, config || {})
    };
  }

  removeFromGroups(itemKeys, groupKeys) {
    var _itemKeys = utils.normalizeArray(itemKeys);

    var _groupKeys = utils.normalizeArray(groupKeys);

    _itemKeys.forEach(itemKey => {
      var removedFromGroupsCount = 0;

      _groupKeys.forEach(groupKey => {
        var group = this.getGroup(groupKey, {
          notExisting: true
        });
        if (!(group !== null && group !== void 0 && group.has(itemKey))) return;
        group.remove(itemKey);
        removedFromGroupsCount++;
      });

      if (removedFromGroupsCount >= this.getGroupKeysThatHaveItemKey(itemKey).length) this.removeItems(itemKey);
    });

    return this;
  }

  removeItems(itemKeys) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      notExisting: false,
      removeSelector: false
    });

    var _itemKeys = utils.normalizeArray(itemKeys);

    _itemKeys.forEach(itemKey => {
      var _item$persistent2;

      var item = this.getItem(itemKey, {
        notExisting: config.notExisting
      });
      if (item == null) return;
      var wasPlaceholder = item.isPlaceholder;

      for (var groupKey in this.groups) {
        var group = this.getGroup(groupKey, {
          notExisting: true
        });
        if (group !== null && group !== void 0 && group.has(itemKey)) group === null || group === void 0 ? void 0 : group.remove(itemKey);
      }

      (_item$persistent2 = item.persistent) === null || _item$persistent2 === void 0 ? void 0 : _item$persistent2.removePersistedValue();
      delete this.data[itemKey];

      for (var selectorKey in this.selectors) {
        var selector = this.getSelector(selectorKey, {
          notExisting: true
        });

        if (selector != null && selector.hasSelected(itemKey, false)) {
          if (config.removeSelector) {
            var _selector$_key;

            this.removeSelector((_selector$_key = selector._key) !== null && _selector$_key !== void 0 ? _selector$_key : 'unknown');
          } else {
            selector.reselect({
              force: true
            });
          }
        }
      }

      if (!wasPlaceholder) this.size--;
    });

    return this;
  }

  assignData(data) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      patch: false,
      background: false
    });

    var _data = utils.copy(data);

    var primaryKey = this.config.primaryKey;

    if (!utils.isValidObject(_data)) {
      logCodeManager.log('1B:03:05', {
        replacers: [this._key]
      });
      return false;
    }

    if (!Object.prototype.hasOwnProperty.call(_data, primaryKey)) {
      logCodeManager.log('1B:02:05', {
        replacers: [this._key, primaryKey]
      });
      _data[primaryKey] = utils.generateId();
    }

    var itemKey = _data[primaryKey];
    var item = this.getItem(itemKey, {
      notExisting: true
    });
    var wasPlaceholder = (item === null || item === void 0 ? void 0 : item.isPlaceholder) || false;

    if (item != null) {
      if (config.patch) {
        item.patch(_data, {
          background: config.background
        });
      } else {
        item.set(_data, {
          background: config.background
        });
      }
    } else {
      this.assignItem(new Item(this, _data), {
        background: config.background
      });
    }

    if (wasPlaceholder) this.size++;
    return true;
  }

  assignItem(item) {
    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};
    config = utils.defineConfig(config, {
      overwrite: false,
      background: false,
      rebuildGroups: true
    });
    var primaryKey = this.config.primaryKey;
    var itemKey = item._value[primaryKey];
    var increaseCollectionSize = true;

    if (!Object.prototype.hasOwnProperty.call(item._value, primaryKey)) {
      logCodeManager.log('1B:02:05', {
        replacers: [this._key, primaryKey]
      });
      itemKey = utils.generateId();
      item.patch({
        [this.config.primaryKey]: itemKey
      }, {
        background: config.background
      });
      item._key = itemKey;
    }

    if (item.collection() !== this) {
      logCodeManager.log('1B:03:06', {
        replacers: [this._key, item.collection()._key]
      });
      return false;
    }

    if (this.getItem(itemKey, {
      notExisting: true
    }) != null) {
      if (!config.overwrite) {
        this.assignData(item._value);
        return true;
      } else increaseCollectionSize = false;
    }

    this.data[itemKey] = item;
    if (config.rebuildGroups) this.rebuildGroupsThatIncludeItemKey(itemKey, {
      background: config.background
    });
    if (increaseCollectionSize) this.size++;
    return true;
  }

  rebuildGroupsThatIncludeItemKey(itemKey) {
    var _this = this;

    var config = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : {};

    var _loop = function _loop(groupKey) {
      var group = _this.getGroup(groupKey);

      if (group != null && group.has(itemKey)) {
        var index = group._preciseItemKeys.findIndex(ik => itemKey === ik);

        if (index !== -1) {
          group.rebuild([{
            key: itemKey,
            index: index,
            method: exports.TrackedChangeMethod.UPDATE
          }], config);
        } else {
          var indexOfBeforeItemKey = group.nextStateValue.findIndex(ik => itemKey === ik) - 1;
          group.rebuild([{
            key: itemKey,
            index: indexOfBeforeItemKey >= 0 ? group._preciseItemKeys.findIndex(ik => group.nextStateValue[indexOfBeforeItemKey] === ik) + 1 : 0,
            method: exports.TrackedChangeMethod.ADD
          }], config);
        }
      }
    };

    for (var groupKey of Object.keys(this.groups)) {
      _loop(groupKey);
    }
  }

}

function createCollection(config) {
  var agileInstance = arguments.length > 1 && arguments[1] !== undefined ? arguments[1] : exports.sharedAgileInstance;
  return new Collection(agileInstance, config);
}

var Agile = Agile$1;

exports.Agile = Agile$1;
exports.CallbackSubscriptionContainer = CallbackSubscriptionContainer;
exports.Collection = Collection;
exports.CollectionPersistent = CollectionPersistent;
exports.ComponentSubscriptionContainer = ComponentSubscriptionContainer;
exports.Computed = Computed;
exports.ComputedTracker = ComputedTracker;
exports.EnhancedState = EnhancedState;
exports.Group = Group;
exports.GroupObserver = GroupObserver;
exports.Integration = Integration;
exports.Integrations = Integrations;
exports.Item = Item;
exports.LogCodeManager = LogCodeManager;
exports.Observer = Observer;
exports.Persistent = Persistent;
exports.Runtime = Runtime;
exports.RuntimeJob = RuntimeJob;
exports.Selector = Selector;
exports.State = State;
exports.StateObserver = StateObserver;
exports.StatePersistent = StatePersistent;
exports.StateRuntimeJob = StateRuntimeJob;
exports.Storage = Storage;
exports.Storages = Storages;
exports.SubController = SubController;
exports.SubscriptionContainer = SubscriptionContainer;
exports.assignAdditionalLogs = assignAdditionalLogs;
exports.assignSharedAgileInstance = assignSharedAgileInstance;
exports.assignSharedStorageManager = assignSharedStorageManager;
exports.createCollection = createCollection;
exports.createComputed = createComputed;
exports.createEnhancedState = createEnhancedState;
exports.createLightState = createLightState;
exports.createState = createState;
exports.createStorage = createStorage;
exports.createStorageManager = createStorageManager;
exports['default'] = Agile;
exports.extractObservers = extractObservers;
exports.extractRelevantObservers = extractRelevantObservers;
exports.getAgileInstance = getAgileInstance;
exports.getSharedStorageManager = getSharedStorageManager;
exports.globalBind = globalBind;
exports.logCodeManager = logCodeManager;
exports.logCodeTypes = logCodeTypes;
exports.optionalRequire = optionalRequire;
exports.runsOnServer = runsOnServer;
exports.shared = exports.sharedAgileInstance;
Object.keys(utils).forEach(function (k) {
  if (k !== 'default' && !exports.hasOwnProperty(k)) Object.defineProperty(exports, k, {
    enumerable: true,
    get: function () {
      return utils[k];
    }
  });
});
//# sourceMappingURL=index.js.map
