import { CreateAgileSubInstanceInterface } from '../shared';
import { CreateStoragesConfigInterface, Storages } from './storages';
/**
 * Handles the permanent persistence of Agile Classes.
 */
declare let sharedStorageManager: Storages | null;
export { sharedStorageManager };
/**
 * Assigns the specified Storage Manager
 * as default (shared) Storage Manager for all Agile Instances.
 *
 *  @param instance - Storage Manager to be assigned as the shared Storage Manager.
 */
export declare const assignSharedStorageManager: (instance: Storages | null) => void;
/**
 * Returns a newly created Storage Manager.
 *
 * A Storage Manager manages all external Storages for AgileTs
 * and provides an interface to easily store,
 * load and remove values from multiple external Storages at once.
 *
 * @param config - Configuration object
 */
export declare function createStorageManager(config?: CreateStorageManagerConfigInterfaceWithAgile): Storages;
/**
 * Returns the shared Storage Manager
 * or creates a new one when no shared Storage Manager exists.
 */
export declare function getSharedStorageManager(): Storages;
export interface CreateStorageManagerConfigInterfaceWithAgile extends CreateAgileSubInstanceInterface, CreateStoragesConfigInterface {
}
